# TODO: refactor to simplify
from pyexlatex.logic.extract.tools import _is_collection


def extract_objs_of_type_from_ambiguous_collection(collection, obj_types):
    collected_objs = []
    if isinstance(collection, dict):
        collected_objs += _extract_objs_of_type_from_dict(collection, obj_types)
    elif _is_collection(collection, obj_types):
        collected_objs += _extract_objs_of_type_from_normal_iterable(collection, obj_types)
    elif isinstance(collection, obj_types):
        collected_objs.append(collection)
    else:
        raise ValueError(f'could not extract objs from {collection} of type {type(collection)}')

    return collected_objs


def _extract_objs_of_type_from_dict(collection, obj_types):
    collected_objs = []
    for key, obj in collection.items():
        if isinstance(collection, dict):
            collected_objs += _extract_objs_of_type_from_dict(collection, obj_types)
        elif _is_collection(obj, obj_types):
            collected_objs += extract_objs_of_type_from_ambiguous_collection(obj, obj_types)
        elif isinstance(obj, obj_types):
            collected_objs.append(obj)
        # else, skip the object

    return collected_objs


def _extract_objs_of_type_from_normal_iterable(collection, obj_types):
    collected_objs = []
    for obj in collection:
        if isinstance(collection, dict):
            collected_objs += _extract_objs_of_type_from_dict(collection, obj_types)
        elif _is_collection(obj, obj_types):
            collected_objs += extract_objs_of_type_from_ambiguous_collection(obj, obj_types)
        elif isinstance(obj, obj_types):
            collected_objs.append(obj)
        # else, skip the object

    return collected_objs





from typing import Sequence, List
from pyexlatex.models.lists.item import ListItem
from pyexlatex.models.lists.base import VerticalFillMixin
from pyexlatex.models.format.text.color.main import TextColor
from pyexlatex.presentation.beamer.overlay.overlay import Overlay
from pyexlatex.presentation.beamer.overlay.until_end import UntilEnd
from pyexlatex.presentation.beamer.overlay.next import NextWithIncrement, NextWithoutIncrement
from pyexlatex.models.containeritem import ContainerItem
from pyexlatex.models.item import ItemBase
from pyexlatex.models.lists.base import can_be_included_directly_in_list


class DimAndRevealListItem(ListItem):
    """
    Single list item which reveals on each slide one-by-one and as the next one reveals, the last one dims.

    Use DimAndRevealListItems to construct.
    """
    is_ListItem = True  # so that passing within lists will treat as a list item

    def __init__(self, contents, dim: bool = True, opacity: float = 0.3, **kwargs):
        self.dim = dim
        dim_ov = Overlay([UntilEnd(NextWithoutIncrement(1))])
        next_ov = Overlay([UntilEnd(NextWithIncrement())])

        if dim:
            contents = TextColor(contents, 'black', opacity=opacity, overlay=dim_ov)

        super().__init__(contents, overlay=next_ov, **kwargs)

    def convert_to_regular_item(self):
        if self.dim:
            contents = self.contents.content  # get from inside TextColor
        else:
            contents = self.contents
        item = ListItem(contents)
        self.__dict__.update(item.__dict__)
        self.__class__ = ListItem
        del self.is_DimAndRevealListItem
        del self.dim


class DimAndRevealListItems(VerticalFillMixin, ContainerItem, ItemBase):
    """
    List items which reveal on each slide one-by-one and as the next one reveals, the last one dims.

    Pass to a list class such as OrderedList or UnorderedList
    """
    name = '<dim and reveal container, should not enter latex output>'
    is_ListBase = True  # so that passing within lists will treat as a list

    def __init__(self, contents: Sequence, dim_last_item: bool = False, opacity: float = 0.3,
                 vertical_fill: bool = False, dim_earlier_items: bool = True, **item_kwargs):
        self.orig_contents = contents
        self.dim_last_item = dim_last_item
        self.dim_earlier_items = dim_earlier_items
        self.opacity = opacity
        self.item_kwargs = item_kwargs
        self.vertical_fill = vertical_fill
        self.add_data_from_content(contents)
        self.contents = self._get_contents()

    def _get_contents(self) -> List[DimAndRevealListItem]:
        output = []
        for item in self.orig_contents:
            if can_be_included_directly_in_list(item):
                output.append(item)
            else:
                output.append(
                    DimAndRevealListItem(
                        item,
                        opacity=self.opacity,
                        dim=self.dim_earlier_items,
                        **self.item_kwargs
                    )
                )

        if not self.dim_last_item:
            output[-1] = DimAndRevealListItem(
                self.orig_contents[-1], dim=False, opacity=self.opacity, **self.item_kwargs
            )
        output = self.vertically_space_content(output)

        return output

    def __str__(self) -> str:
        from pyexlatex.logic.builder import _build
        if isinstance(self.contents, str):
            return self.contents
        return _build(self.contents)


def eliminate_dim_reveal(content):
    """
    Eliminates dim/reveal from nested content. Modifies in place by using regular list items
    """
    if hasattr(content, 'content'):
        eliminate_dim_reveal(content.content)
    if hasattr(content, 'contents'):
        eliminate_dim_reveal(content.contents)
    if hasattr(content, 'is_DimAndRevealListItem') and content.is_DimAndRevealListItem:
        content.convert_to_regular_item()
    if isinstance(content, (list, tuple)):
        [eliminate_dim_reveal(c) for c in content]



from pyexlatex.models.item import Item

class FileContents(Item):
    name = 'filecontents'

    def __init__(self, contents, filename: str):
        super().__init__(self.name, contents, env_modifiers=f'{{{filename}}}')


from pyexlatex.models.references.bibtex.style.style import Style

BUILTIN_STYLE_NAMES = [
    'abbrv',
    'acm',
    'alpha',
    'apalike',
    'ieeetr',
    'plain',
    'siam',
    'unsrt',
]

BUILTIN_STYLES = {style_str: Style(style_str) for style_str in BUILTIN_STYLE_NAMES}


from typing import Union, AnyStr, List, TYPE_CHECKING, Optional
import pandas as pd

if TYPE_CHECKING:
    from pyexlatex.table.models.texgen.items import ColumnsAlignment

from mixins.repr import ReprMixin
from pyexlatex.table.models.panels.collection import PanelCollection, Panel
from pyexlatex.table.models.labels.table import LabelTable, LabelCollection
from pyexlatex.table.models.table.caption import Caption
from pyexlatex.logic.pdf.main import document_to_pdf_and_move
from pyexlatex.texgen.replacements.filename import latex_filename_replacements
from pyexlatex.models.documentitem import DocumentItem
from pyexlatex.models.commands.newenvironment import NewEnvironment
from pyexlatex.models.commands.begin import Begin
from pyexlatex.models.commands.end import End
from pyexlatex.models.documentsetup import DocumentSetupData



class Table(DocumentItem, ReprMixin):
    """
    An object for creating latex tables. Easiest way to create is with Table.from_list_of_lists_of_dfs,
    but for more control, construct Panel objects and use Table.from_panel_list
    """
    repr_cols = ['caption', 'above_text', 'panels', 'below_text']

    def __init__(self, panels: PanelCollection, caption: str=None, above_text: str=None, below_text: str=None,
                 align: str = None, mid_rules=True, landscape=False, label: Optional[str] = None):
        """

        :param panels: list of Panels
        :param caption: overall caption to place at top of table
        :param above_text: Not yet implemented
        :param below_text: text to place below table
        :param align: Can take any string that would normally used in tabular (i.e. rrr for three columns right aligned
                        as well as L{<width>), C{<width>}, and R{<width>} (i.e. L{3cm}) for left, center, and right aligned
                        fixed width. Default is first column left aligned, rest center aligned.
        :param mid_rules: whether to add mid rules between panels
        :param landscape: whether to output landscape tex
        :param label: label for table to be referenced in text
        """
        from pyexlatex.table.models.texgen.items import TableNotes
        self.panels = panels
        self.caption = Caption(caption)
        self.above_text = _set_above_text(above_text)
        self.below_text = TableNotes(below_text) if below_text is not None else None
        self.align = align
        self.mid_rules = mid_rules
        self.landscape = landscape
        self.label = label
        self.data = DocumentSetupData()
        self.set_begin_document_items(landscape)

    def __str__(self):
        from pyexlatex.logic.builder import build
        tex_obj = self.tex_obj(as_document=False)

        # TODO: restructure lt module so that it works like others where items are tex generators
        # TODO: until then, need to manually call build as it won't carry through to the separate tex generator
        build(tex_obj)

        return str(tex_obj)

    def to_tex(self, as_document=True):
        tex_generator = self.tex_obj(as_document=as_document)
        return str(tex_generator)

    def tex_obj(self, as_document=True):
        from pyexlatex.table.models.texgen.items import TableDocument, Table as TexTable, LTable
        if as_document:
            class_factory = TableDocument.from_table_model
        else:
            if self.landscape:
                class_factory = LTable.from_table_model
            else:
                class_factory = TexTable.from_table_model

        tex_generator = class_factory(self)
        return tex_generator

    def to_pdf_and_move(self, outfolder, outname=None,
                              move_folder_name='Tables', as_document=True):
        tex = self.to_tex(as_document=as_document)

        if outname is None:
            outname = self.caption

        outname = latex_filename_replacements(outname)

        document_to_pdf_and_move(
            tex,
            outfolder=outfolder,
            outname=outname,
            move_folder_name=move_folder_name,
            as_document=as_document
        )

    @classmethod
    def from_panel_list(cls, panels: [Panel], label_consolidation: str='object', enforce_label_order=True,
                        top_left_corner_labels: Union[LabelTable, LabelCollection, List[AnyStr], AnyStr] = None,
                        pad_rows: int=1, pad_columns: int=1, caption: str=None, above_text: str=None,
                        below_text: str=None, align: str = None, mid_rules=True, landscape=False,
                        label: Optional[str] = None):
        """
        Note: to quickly create latex tables, use Table.from_list_of_lists_of_dfs to pass
        pandas DataFrames directly. Use this method when more control is needed than is provided by that method.
        Construct Panels individually, either directly from DataFrames with Panel.from_df_list, or for even more
        control, use DataTable.from_df and override labels with custom LabelTables.

        :param panels: list of Panels
        :param label_consolidation: pass 'object' to to avoid consolidating labels. pass 'str' to consolidate if
                                    column or index contents are identical within a column or row of DataFrames
        :param enforce_label_order: pass False to allow consolidating lower labels even if upper labels do not match.
                                    e.g. if labels on one table are [['Top1'], ['Bot1', 'Bot2']], then labels on the other
                                    table are [['Top2'], ['Bot1', 'Bot2']], consolidated labels when passing False will be
                                    ['Bot1', 'Bot2'], while when passing True, no labels will be consolidated. Under True,
                                    will start from the top label, then stop consolidating once it has a mismatch.
        :param top_left_corner_labels: additional labels to place in the top left corner. pass a single string
                                       or a list of strings for convenience. a list of strings will be create labels
                                       which span the gap horizontally and go downwards, one label per row. pass
                                       LabelCollection or LabelTable for more control.
        :param pad_rows: horizontal spacing to put between panels
        :param pad_columns: vertical spacing to put between TableSections
        :param caption: overall caption to place at top of table
        :param above_text: Not yet implemented
        :param below_text: text to place below table
        :param align: Can take any string that would normally used in tabular (i.e. rrr for three columns right aligned
                        as well as L{<width>), C{<width>}, and R{<width>} (i.e. L{3cm}) for left, center, and right aligned
                        fixed width. Default is first column left aligned, rest center aligned.
        :param mid_rules: whether to add mid rules between panels
        :param landscape: whether to output landscape tex
        :param label: label for table to be referenced in text
        :return:
        """
        panel_collection = PanelCollection(
            panels,
            label_consolidation=label_consolidation,
            enforce_label_order=enforce_label_order,
            top_left_corner_labels=top_left_corner_labels,
            pad_rows=pad_rows,
            pad_columns=pad_columns,
            name=caption
        )

        return cls(
            panel_collection,
            caption=caption,
            above_text=above_text,
            below_text=below_text,
            align=align,
            mid_rules=mid_rules,
            landscape=landscape,
            label=label
        )

    @classmethod
    def from_list_of_lists_of_dfs(cls, df_list_of_lists: [[pd.DataFrame]],
                                  panel_names: [str]=None, shape: tuple=None,
                                  include_columns=True, include_index=False,
                                  label_consolidation: str = 'str', enforce_label_order=True,
                                  top_left_corner_labels: Union[LabelTable, LabelCollection, List[AnyStr], AnyStr] = None,
                                  pad_rows: int = 1, pad_columns: int = 1, caption: str = None, above_text: str = None,
                                  below_text: str = None, align: str = None, mid_rules=True, landscape=False,
                                  label: Optional[str] = None,
                                  data_table_kwargs={}
                                  ):
        """
        To create a single panel table, pass a single list within
        a list of DataFrames, e.g. [[df1, df2]] then shape will specify how the DataFrames will
        be organized in the Panel. If you pass two lists within the outer list, then shape will
        apply to each Panel. So [[df1, df2], [df3, df4]] with shape=(1,2) create a two Panel table
        with two tables placed within each panel going horizontally, so that the overall shape is (2,2).

        Note: convenience method for if not much control over table is needed.
        To apply different options to each panel, construct them individually using
        Panel.from_df_list or sub panels individually with DataTable.from_df then make modifications

        :param df_list_of_lists: list of lists of pandas DataFrame
        :param panel_names: list of panel names. Must be of same length as outer list in df_list_of_lists
        :param shape: tuple of (rows, columns) to arrange DataFrames. They will be placed from left to right,
                      then from top to bottom.
                      passsing None defaults one column, as many rows as DataFrames.
                      Note: this is for each panel. To apply a different shape to each Panel, construct Panels individually
                      and pass to Table.from_panel_list
        :param include_columns:
        :param include_index:
        :param label_consolidation: pass 'object' to to avoid consolidating labels. pass 'str' to consolidate if
                                    column or index contents are identical within a column or row of DataFrames
        :param top_left_corner_labels: additional labels to place in the top left corner. pass a single string
                                       or a list of strings for convenience. a list of strings will be create labels
                                       which span the gap horizontally and go downwards, one label per row. pass
                                       LabelCollection or LabelTable for more control.
        :param enforce_label_order: pass False to allow consolidating lower labels even if upper labels do not match.
                                    e.g. if labels on one table are [['Top1'], ['Bot1', 'Bot2']], then labels on the other
                                    table are [['Top2'], ['Bot1', 'Bot2']], consolidated labels when passing False will be
                                    ['Bot1', 'Bot2'], while when passing True, no labels will be consolidated. Under True,
                                    will start from the top label, then stop consolidating once it has a mismatch.
        :param pad_rows: horizontal spacing to put between panels
        :param pad_columns: vertical spacing to put between TableSections
        :param caption: overall caption to place at top of table
        :param above_text: Not yet implemented
        :param below_text: text to place below table
        :param align: Can take any string that would normally used in tabular (i.e. rrr for three columns right aligned
                        as well as L{<width>), C{<width>}, and R{<width>} (i.e. L{3cm}) for left, center, and right aligned
                        fixed width. Default is first column left aligned, rest center aligned.
        :param mid_rules: whether to add mid rules between panels
        :param landscape: whether to output landscape tex
        :param label: label for table to be referenced in text
        :param data_table_kwargs: kwargs to be passed to DataTable.from_df. Same kwargs will be passed to
                                  all data tables.
        :return:
        """
        panel_collection = PanelCollection.from_list_of_lists_of_dfs(
            df_list_of_lists,
            panel_names=panel_names,
            panel_kwargs=dict(
                shape=shape,
                include_columns=include_columns,
                include_index=include_index,
            ),
            label_consolidation=label_consolidation,
            enforce_label_order=enforce_label_order,
            top_left_corner_labels=top_left_corner_labels,
            pad_rows=pad_rows,
            pad_columns=pad_columns,
            name=caption,
            data_table_kwargs=data_table_kwargs
        )

        return cls(
            panel_collection,
            caption=caption,
            above_text=above_text,
            below_text=below_text,
            align=align,
            mid_rules=mid_rules,
            landscape=landscape,
            label=label
        )

    @classmethod
    def from_panel_name_df_dict(cls, panel_name_df_dict: {str: pd.DataFrame},
                                include_columns=True, include_index=False,
                                label_consolidation: str = 'str', enforce_label_order=True,
                                top_left_corner_labels: Union[LabelTable, LabelCollection, List[AnyStr], AnyStr] = None,
                                pad_rows: int = 1, caption: str = None, above_text: str = None,
                                below_text: str = None, align: str = None, mid_rules=True, landscape=False,
                                label: Optional[str] = None,
                                data_table_kwargs={}
                                ):
        """
        Convenience method for when there is only one DataFrame per panel. All options will be applied all panels and
        DataTables.
        To apply different options to each panel, construct them individually using
        Panel.from_df_list or sub panels individually with DataTable.from_df then make modifications

        :param panel_name_df_dict: dictionary where keys are names of panels, and values are DataFrames to put in panels
                                   (only one DataFrame per panel. Use another method if more control is needed)
        :param include_columns:
        :param include_index:
        :param label_consolidation: pass 'object' to to avoid consolidating labels. pass 'str' to consolidate if
                                    column or index contents are identical within a column or row of DataFrames
        :param top_left_corner_labels: additional labels to place in the top left corner. pass a single string
                                       or a list of strings for convenience. a list of strings will be create labels
                                       which span the gap horizontally and go downwards, one label per row. pass
                                       LabelCollection or LabelTable for more control.
        :param enforce_label_order: pass False to allow consolidating lower labels even if upper labels do not match.
                                    e.g. if labels on one table are [['Top1'], ['Bot1', 'Bot2']], then labels on the other
                                    table are [['Top2'], ['Bot1', 'Bot2']], consolidated labels when passing False will be
                                    ['Bot1', 'Bot2'], while when passing True, no labels will be consolidated. Under True,
                                    will start from the top label, then stop consolidating once it has a mismatch.
        :param pad_rows: horizontal spacing to put between panels
        :param caption: overall caption to place at top of table
        :param above_text: Not yet implemented
        :param below_text: text to place below table
        :param align: Can take any string that would normally used in tabular (i.e. rrr for three columns right aligned
                        as well as L{<width>), C{<width>}, and R{<width>} (i.e. L{3cm}) for left, center, and right aligned
                        fixed width. Default is first column left aligned, rest center aligned.
        :param mid_rules: whether to add mid rules between panels
        :param landscape: whether to output landscape tex
        :param label: label for table to be referenced in text
        :param data_table_kwargs: kwargs to be passed to DataTable.from_df. Same kwargs will be passed to
                                  all data tables.
        :return:
        """

        df_list_of_lists = []
        panel_names = []
        for panel_name, df in panel_name_df_dict.items():
            df_list_of_lists.append([df])
            panel_names.append(panel_name)

        return cls.from_list_of_lists_of_dfs(
            df_list_of_lists,
            panel_names=panel_names,
            include_columns=include_columns,
            include_index=include_index,
            label_consolidation=label_consolidation,
            enforce_label_order=enforce_label_order,
            top_left_corner_labels=top_left_corner_labels,
            pad_rows=pad_rows,
            caption=caption,
            above_text=above_text,
            below_text=below_text,
            align=align,
            mid_rules=mid_rules,
            landscape=landscape,
            data_table_kwargs=data_table_kwargs,
            label=label
        )

    @property
    def align(self):
        return self._align

    @align.setter
    def align(self, align: Union['ColumnsAlignment', str]):
        from pyexlatex.table.models.texgen.items import ColumnsAlignment
        if align is None:
            self._align = ColumnsAlignment(num_columns=self.panels.num_columns)
        elif isinstance(align, ColumnsAlignment):
            self._align = align
        elif isinstance(align, str):
            self._align = ColumnsAlignment.from_alignment_str(align, num_columns=self.panels.num_columns)
        else:
            raise NotImplementedError(f'could not create align from {align}')

    def set_begin_document_items(self, landscape: bool):
        begin_doc_items = []
        if landscape:
            ltable_def = NewEnvironment(
                'ltable',
                Begin('landscape') + Begin('table'),
                End('table') + End('landscape')
            )
            begin_doc_items.append(ltable_def)
        self.data.begin_document_items.extend(begin_doc_items) 





def _set_above_text(above_text):
    # TODO: handle above text
    if above_text is not None:
        raise NotImplementedError('will add above text in a future release')



from pyexlatex.models.control.documentclass.classtypes.builtin_classes import BUILTIN_CLASS_TYPES
from pyexlatex.models.control.documentclass.classtypes.custom import CUSTOM_CLASS_TYPES
from pyexlatex.models.control.documentclass.classtypes.documentclasstype import DocumentClassType


class DocumentClassTypesManager:

    def __init__(self):
        pass

    def get_class_type_by_name(self, name: str) -> DocumentClassType:
        if self.is_builtin_class_type(name):
            return BUILTIN_CLASS_TYPES[name]

        if self.is_custom_class_type(name):
            return CUSTOM_CLASS_TYPES[name]

        raise NoSuchDocumentClassTypeException

    def is_custom_class_type(self, name: str) -> bool:
        return name in CUSTOM_CLASS_TYPES

    def is_builtin_class_type(self, name: str) -> bool:
        return name in BUILTIN_CLASS_TYPES


class NoSuchDocumentClassTypeException(Exception):
    pass



from typing import Union, List, Optional

import pandas as pd

from pyexlatex.table.models.interfaces import LabelClassOrStrs
from pyexlatex.logic.tools import _add_if_not_none
from mixins.repr import ReprMixin
from pyexlatex.table.models.data.valuestable import ValuesTable
from pyexlatex.table.models.labels.label import Label
from pyexlatex.table.models.labels.table import LabelTable, LabelCollection
from pyexlatex.table.models.spacing.columntable import ColumnPadTable, CellSpacer
from pyexlatex.table.models.table.section import TableSection
from pyexlatex.table.logic.panels.topleft import _set_top_left_corner_labels


class DataTable(TableSection, ReprMixin):
    """
    Represents a subsection in a panel, but tracks row and column labels, which may be consolidated when
    assembled into Panels then a Table

    Use DataTable.from_df to create a DataTable from a pandas DataFrame.
    """
    repr_cols = ['values_table', 'column_labels', 'row_labels']

    def __init__(self, values_table: ValuesTable, column_labels: LabelTable=None, row_labels: LabelTable=None,
                 top_left_corner_labels: Union[Label, str] = None, break_size_adjustment: Optional[str] = None):
        self.values_table = values_table
        self.column_labels = column_labels
        self.row_labels = row_labels
        self.top_left_corner_labels = _set_top_left_corner_labels(top_left_corner_labels)
        self.break_size_adjustment = break_size_adjustment

        self.should_add_top_left = (self.has_column_labels) and (self.has_row_labels)

    def __add__(self, other):
        if isinstance(other, DataTable):
            row_labels_match = _determine_match(self.row_labels, other.row_labels)
            if row_labels_match or other.row_labels is None:
                # if right table has same or None row labels, eliminate right row labels. Just add values
                values_table = _add_if_not_none(self.values_table, other.values_table)
                column_labels = _add_if_not_none(self.column_labels, other.column_labels)
            else:
                # if right table has unique row labels, absorb them into middle of values table
                values_table = _add_if_not_none(
                    self.values_table,
                    ValuesTable(other.row_labels.rows),
                    other.values_table
                )
                column_labels = _add_if_not_none(
                    self.column_labels,
                    other.top_left_corner_labels,
                    other.column_labels
                )

            row_labels = self.row_labels
        elif isinstance(other, ColumnPadTable):
            values_table = self.values_table + other if self.values_table is not None else None
            column_labels = self.column_labels + other if self.column_labels is not None else None
            row_labels = self.row_labels
        elif isinstance(other, TableSection):
            values_table = _add_if_not_none(self.values_table, other)
            column_labels = self.column_labels
            row_labels = self.row_labels
        else:
            raise ValueError(f'must add DataTable or TableSection to type {type(self)}. Got type {type(other)}')

        return DataTable(
            values_table=values_table,
            column_labels=column_labels,
            row_labels=row_labels,
            top_left_corner_labels=self.top_left_corner_labels
        )

    @property
    def rows(self):
        try:
            return self._rows
        except AttributeError:
            self._rows = self._create_rows()

        return self._rows

    # Following properties/setters exist to recreate rows if user overrides values table or labels

    @property
    def column_labels(self):
        return self._column_labels

    @column_labels.setter
    def column_labels(self, labels: LabelTable):
        self._column_labels = labels
        self._recreate_rows_if_created()

    @property
    def has_column_labels(self) -> bool:
        return (self.column_labels is not None and not self.column_labels.is_empty)

    @property
    def row_labels(self):
        return self._row_labels

    @row_labels.setter
    def row_labels(self, labels: LabelTable):
        self._row_labels = labels
        self._recreate_rows_if_created()

    @property
    def has_row_labels(self) -> bool:
        return (self.row_labels is not None and not self.row_labels.is_empty)

    @property
    def values_table(self):
        return self._values_table

    @values_table.setter
    def values_table(self, table: ValuesTable):
        self._values_table = table
        self._recreate_rows_if_created()

    def _create_rows(self):

        rows = []
        if self.has_column_labels:
            if self.should_add_top_left:
                column_labels = self.top_left_corner_labels + self.column_labels
            else:
                column_labels = self.column_labels
            rows += column_labels.rows

        # need to add row labels inline with values table
        if self.has_row_labels:
            assert len(self.row_labels.rows) == len(self.values_table.rows)
            for label_row, value_row in zip(self.row_labels.rows, self.values_table.rows):
                rows.append(label_row + value_row)
        # no row labels, add values rows as they are
        else:
            rows += self.values_table.rows

        return rows

    @classmethod
    def from_df(cls, df: pd.DataFrame, include_columns=True, include_index=False,
                extra_header: str=None, extra_header_underline=True,
                top_left_corner_labels: LabelClassOrStrs = None,
                **kwargs):
        """
        Use for the most fine-grained control in creating tables. Construct DataTables from
        pandas DataFrames, modify labels as needed, assemble them into Panels, then create a latex Table with
        Table.from_panel_list.

        :param df:
        :param include_columns:
        :param include_index:
        :param extra_header: extra headers to place over the existing column labels (or over values if
                             there are no column labels). Useful when placing multiple DataTables horizontally
                             in a Panel. If a str is passed, str will become a multicolumn header for the entire
                             DataTable. If a list of as many strs as there are columns is passed, these
                             will be placed above any existing column labels. LabelTable and LabelCollection objects
                             may also be passed for more control. When those objects are passed, the extra_header_underline
                             argument is ignored.
        :param extra_header_underline: whether to add an underline under the extra header, if the extre header
                                       was passed
        :param top_left_corner_labels: additional labels to place in the top left corner. pass a single string
                                       or a list of strings for convenience. a list of strings will be create labels
                                       which span the gap horizontally and go downwards, one label per row. pass
                                       LabelCollection or LabelTable for more control.
        :param kwargs: DataTable kwargs
        :return:
        """
        values_table = ValuesTable.from_df(df)

        if include_columns:
            column_label_table = LabelTable.from_df_index(df.columns)
        else:
            column_label_table = None

        if include_index:
            row_label_table = LabelTable.from_df_index(df.index).T
        else:
            row_label_table = None

        dt = cls(
            values_table,
            column_labels=column_label_table,
            row_labels=row_label_table,
            top_left_corner_labels=top_left_corner_labels,
            **kwargs,
        )

        if extra_header is not None:

            header = _create_header_label_collection_list(extra_header, values_table, extra_header_underline)

            if include_columns:
                # add to existing
                header.reverse() # need to insert end first, so they end up in original order
                [dt.column_labels.label_collections.insert(0, label_collection) for label_collection in header]
            else:
                # create column labels as extra header
                dt.column_labels = LabelTable(header)

            if include_index:  # need to add to top left
                num_spacers = len(header)

                # top left has one by default. if there were already columns, that top left is used up by the columns
                # and we need to add more. if there are not columns, then this one existing top left can be used
                # for the new header
                if not include_columns:
                    num_spacers -= 1

                for i in range(num_spacers):
                    dt.top_left_corner_labels.label_collections.insert(0, LabelCollection.from_str_list([' ']))

            # add spacer to top left corner labels to match new row of headers
            if (not include_columns) and include_index:
                # header becomes new columns, have to add top left as if there were columns
                dt.should_add_top_left = True


        return dt

def _determine_match(labels1: LabelTable, labels2: LabelTable):
    # handle equality for None
    if labels1 is None:
        if labels2 is None:
            return True
        else:
            return False
    elif labels2 is None:
        # labels 1 must not be None here
        return False

    # here, both are not None
    return labels1.matches(labels2)

def _create_header_label_collection_list(extra_header: LabelClassOrStrs, values_table: ValuesTable,
                                         underline: bool) -> List[LabelCollection]:

    if isinstance(extra_header, LabelTable):
        return extra_header.label_collections

    if isinstance(extra_header, LabelCollection):
        return [extra_header]

    if isinstance(extra_header, list):
        if len(extra_header) != values_table.num_columns:
            raise ValueError(f'passed extra header has {len(extra_header)} columns, while table has '
                             f'{values_table.num_columns} columns')
        if underline:
            underline_arg = f'0-{len(extra_header) - 1}'
        else:
            underline_arg = None
        return [LabelCollection.from_str_list(extra_header, underline=underline_arg)]

    # Got a string or latex item

    # create multicolumn label
    label = Label(extra_header, span=values_table.num_columns)
    # set underline
    if underline:
        underline_arg = 0  # place an underline under the singular label
    else:
        underline_arg = None  # no underline
    return [LabelCollection([label], underline=underline_arg)]


from pyexlatex.texgen import _usepackage_str
from pyexlatex.models.item import ItemBase


class Package(ItemBase):
    r"""
    Represents LaTeX \usepackage{}, pass to Document if any custom LaTeX packages are needed.
    """
    equal_attrs = ['name', 'modifier_str']

    def __init__(self, name, modifier_str=None):
        self.name = name
        self.modifier_str = modifier_str

    def __repr__(self):
        return f'<Package(name={self.name})>'

    def __str__(self):
        return _usepackage_str(self.name, self.modifier_str)

    def matches_name(self, other):
        return self.name == other




from typing import Sequence, Optional
from pyexlatex.models.item import ItemBase
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.presentation.beamer.overlay.overlay import Overlay
from pyexlatex.presentation.beamer.overlay.commands.uncover import Uncover


class TikzOptionHandler:
    options = None

    @property
    def options_str(self) -> str:
        if self.options is None:
            return ''
        str_options = [str(option) for option in self.options]
        return self._wrap_with_bracket(', '.join(str_options))

    def _wrap_with_bracket(self, to_wrap: str):
        raise NotImplementedError('must also subclass ItemBase')


class TikZItem(TextAreaMixin, ItemBase, TikzOptionHandler):

    def __init__(self, name, contents, options: Optional[Sequence[str]] = None,
                 overlay: Optional['Overlay'] = None):
        self.contents = contents
        self.options = options
        self.overlay = overlay
        self.name = name
        super().__init__(name, contents)

    def __str__(self):
        overlay = self.overlay if self.overlay is not None else ""
        item_str = fr'\{self.name}{overlay} {self.options_str} {self.contents};'
        return item_str








from pyexlatex.models.item import SimpleItem


class NewCounter(SimpleItem):
    name = 'newcounter'

    def __init__(self, counter_name: str, **kwargs):
        self.counter_name = counter_name

        super().__init__(self.name, self.counter_name, **kwargs)



import re
from copy import deepcopy
from functools import partial
from typing import Sequence, List, Any, Callable

from jinja2 import Environment, Template

from pyexlatex.models.containeritem import ContainerItem
from pyexlatex.models.datastore import DataStore
from pyexlatex.models.documentsetup import DocumentSetupData
from pyexlatex.models.item import ItemBase
from pyexlatex.texgen.replacements.file import general_latex_replacements

UPPER_PATTERN = re.compile('[A-Z]')

current_template_data_store = None


def class_factory(latex_class, *args, **kwargs):
    item = latex_class(*args, **kwargs)
    global current_template_data_store
    current_template_data_store.add_data_from_content(item)
    return str(item)


def get_capitalized_items(items: Sequence[str]) -> List[str]:
    return [name for name in items if UPPER_PATTERN.match(name[0])]


class JinjaTemplate(Template, ContainerItem):
    """
    A jinja Template but with pyexlatex models as built-in filters and handling extracting pyexlatex data

    Examples:

        >>> import pyexlatex as pl
        >>> str(pl.JinjaTemplate('{{ my_var | Italics }}').render(my_var='woo'))
        '\\textit{woo}'
    """

    def __new__(cls, source, **kwargs):
        env = JinjaEnvironment(**kwargs)
        return env.from_string(source, template_class=cls)

    def __init__(self, *args, **kwargs):
        pass

    def render(self, *args, **kwargs):
        format_dict = dict(*args, **kwargs)
        self.add_data_from_content(format_dict)

        # Set as current global template for adding data during filters
        _set_data_store_to_object(self)

        string = super().render(*args, **kwargs)
        return DataString(string, self.data)

    def __deepcopy__(self, memo):
        # TODO: may be able to remove once https://github.com/pallets/jinja/issues/758 is resolved
        # Boilerplate deepcopy
        cls = self.__class__

        # The one modification to boilerplate deepcopy, originally cls and not object
        # Create instance without using JinjaTemplate.__new__
        # This is the way it is being done in Template._from_namespace and it avoids an error during
        # deepcopy that source is not defined
        result = object.__new__(cls)

        # Continue boilerplate deepcopy
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result



class DataString(ItemBase):

    def __init__(self, string: str, data: DocumentSetupData):
        super().__init__()
        self.content = string
        self.data = data

    def __str__(self):
        return general_latex_replacements(self.content)


class JinjaEnvironment(Environment):
    """
    A jinja Environment but with pyexlatex models as built in filters and handling extracting pyexlatex data

    Examples:

        >>> import pyexlatex as pl
        >>> from jinja2 import DictLoader
        >>> env = pl.JinjaEnvironment(
        >>>     loader=DictLoader({'my_temp': '{{ my_var | Italics }}'})
        >>> )
        >>> temp = env.get_template('my_temp')
        >>> str(temp.render(my_var='woo'))
        '\\textit{woo}'

    """
    template_class = JinjaTemplate

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._add_filters()

    def _add_filters(self):
        import pyexlatex as pl
        import pyexlatex.table as lt
        import pyexlatex.presentation as lp
        import pyexlatex.graphics as lg
        import pyexlatex.layouts as ll
        import pyexlatex.figure as lf

        for module in [pl, lt, lp, lg, ll, lf]:
            latex_class_names = get_capitalized_items(dir(module))
            for latex_class_name in latex_class_names:
                latex_class = getattr(module, latex_class_name)
                this_class_factory = partial(class_factory, latex_class)
                self.filters[latex_class_name] = this_class_factory

    def from_string(self, *args, **kwargs):
        return _template_factory(
            super().from_string,
            *args,
            **kwargs
        )

    def _load_template(self, *args, **kwargs):
        return _template_factory(
            super()._load_template,
            *args,
            **kwargs
        )


def _set_data_store_to_temporary_object():
    """
    Capture data by using a temporary object
    :return:
    """
    temp_obj = DataStore()
    _set_data_store_to_object(temp_obj)
    return temp_obj


def _set_data_store_to_object(obj: Any):
    global current_template_data_store
    current_template_data_store = obj


def _template_factory(factory_func: Callable, *args, **kwargs) -> JinjaTemplate:
    """
    Handles creating jinja template complete with pyexlatex data.

    Manages a temporary global data store so that jinja filters can add to that data store, then after creating
    the template, the data is added to the template.

    :param factory_func: function which should return a Template
    :param args: passed to factory_func
    :param kwargs: passed to factory_func
    """
    # Set current global data store for adding data during filters
    data_store = _set_data_store_to_temporary_object()

    # Create object in usual jinja way
    actual_template = factory_func(*args, **kwargs)

    # Add data to newly created object
    actual_template.data = data_store.data

    return actual_template




from typing import Dict
from pyexlatex.texgen import bibtex_str
from pyexlatex.models.documentitem import DocumentItem
from mixins.repr import ReprMixin

class BibTexEntryBase(DocumentItem, ReprMixin):
    is_BibTexEntry = True
    item_type = 'notimplemented'
    repr_cols = ['item_accessor', 'fields']
    required_attrs = []
    optional_attrs = []

    def __init__(self, item_accessor: str):
        self.item_accessor = item_accessor

        super().__init__()

    def __str__(self):
        str_ = bibtex_str(
            self.item_type,
            self.item_accessor,
            self.fields
        )

        return str_

    @property
    def fields(self) -> Dict[str, str]:
        fields_dict = {arg: getattr(self, arg) for arg in self.required_attrs}

        for attr in self.optional_attrs:
            self_attr = getattr(self, attr)
            if self_attr:
                fields_dict.update({attr: self_attr})

        return fields_dict


import os

import pandas as pd

from pyexlatex.tools import csv_to_raw_latex, date_time_move_latex


def df_to_pdf_and_move(dflist, outfolder, outname='table', tabular_string='', string_format='',
                       above_text='', below_text='',
                     font_size=12, caption='', missing_rep=' - ', landscape=False,
                       panel_names=None, colname_flags=None,
                       as_document=True, outmethod='pandas'):
    '''
    Takes a dataframe or list of dataframes as input and outputs to a LaTeX formatted table with multiple panels,
    creates a PDF, and moves the LaTeX file and PDF to a dated folder.

    Required options:
        dflist:         Dataframe or list of dataframes.
        outfolder:      Output folder for LaTeX file and PDF. Inside of this folder, a folder called Tables will be created,
                        inside of which the two files will be put inside another folder with the date.


    Optional options:
        outname:        Name of output table, default is table
        tabular_string: Can take any string that would normally used in tabular (i.e. rrr for three columns right aligned
                        as well as L{<width>), C{<width>}, and R{<width>} (i.e. L{3cm}) for left, center, and right aligned
                        fixed width. Additionally . aligns on the decimal. Default is first column left aligned, rest
                        center aligned.
        string_format:  String or list of format of numbers in the table. Please see Python number formats. Pass a blank
                        string to leave formatting untouched (the default).
        above_text:     String of text to display above table
        below_text:     String of text to display below table
        font_size:      Font size, default 12
        caption:        Title of table
        missing_rep:    Representation for missing numbers, default " - "
        landscape:      Boolean. True to switch table to landscape output instead of portrait.
        panel_names:    Python list of names of each panel, to go below column names, e.g. ['Table','Other Table']
        colname_flags:  Python list of yes or no flags for whether to display column names for each panel. Default is to
                        display column names only for the first panel, as usually the panels have the same columns.
                        The default input for a three panel table would be ['y','n','n']
        as_document:    Boolean. True to output latex wrappers for table to be a standalone document. False to write
                        only table wrappers so that table can be included in another document
        outmethod:      String, 'pandas' or 'csv'. If 'pandas', uses pandas' built in df.to_latex() to build latex. If
                        'csv', uses df.to_csv() and then pyexlatex.df.raw_csv_to_latex(). The latter case is useful when the table
                        itself contains latex expressions.

    '''
    from pyexlatex.texgen.replacements.file import general_latex_replacements

    if isinstance(dflist, pd.DataFrame):
        dflist = [dflist]
    assert isinstance(dflist, list)
    if isinstance(string_format, str):
        string_format = [string_format] * len(dflist)
    assert isinstance(string_format, list)

    def is_number(s):
        try:
            float(s)
            return True
        except (ValueError, TypeError):
            return False

    def latex_filename_replacements(string):
        return string.replace('%', 'pct').replace('/','_').replace('\\','_')

    def all_latex_replacements(*tuple):
        return [general_latex_replacements(item) for item in tuple]

    # Latex string replacements will be made in the data below. Here make adjustments to titles, above/below text, etc.
    caption, above_text, below_text = all_latex_replacements(caption, above_text, below_text)
    outname = latex_filename_replacements(outname)
    if panel_names is not None:
        panel_names = all_latex_replacements(*panel_names)

    outname_tex = str(outname) + ".tex"
    outpath = os.path.join(outfolder, outname_tex)
    latex_string_list = [] #set container for final LaTeX table contents
    if (colname_flags is None) or (len(colname_flags) is not len(dflist)): #if the user didn't specify whether to use colnames, or they specified an incorrect number of flags
        colname_flags = ['y'] #set first colnames to show
        for i in range(len(dflist) - 1):
            colname_flags.append('n') #set rest of colnames not to show
    panel_order = -1
    for i, df in enumerate(dflist): #for each csv in the list
        if string_format[i]:
            df = dflist[i].applymap(lambda x: string_format[i].format(float(x)) if is_number(x) else x)
        df = df.fillna(missing_rep)
        if outmethod.lower() == 'pandas':
            latex_list = [line for line in df.to_latex().split('\n') if not line.startswith('\\')]
        elif outmethod.lower() == 'csv':
            latex_list = [line for line in csv_to_raw_latex(df.to_csv(), missing_rep=missing_rep,
                                                            csvstring=True, skipfix='_',
                                                            formatstr=string_format[i]) if not line.startswith('\\')]
        number_of_columns = 1 + latex_list[0].count(' & ') #number of columns is 1 + number of seperators
        if colname_flags[i].lower() in ('n','no'): #if the flag for colnames is no for this panel
            latex_list = latex_list[1:] #chop off colnames
        if panel_names is not None and panel_names[i]:
            panel_order += 1 #In combination with next line, sets panel to A, etc.
            panel_letter = chr(panel_order + ord('A')) #sets first panel to A, second to B, and so on
            #LaTeX formatting code
            panel_latex_list = [
                r'\midrule \\[-11pt]',
                '\n',
                r'\multicolumn{' + str(number_of_columns) + '}{c}{Panel '+ panel_letter + ': ' + panel_names[i] + '} \\\ \\\[-11pt]',
                '\n',
                r'\midrule',
                '\n'
            ]
        else: #if there is no panel name, just put in a midrule
            panel_latex_list = [
                r'\midrule',
                '\n'
            ]
        latex_list = panel_latex_list + latex_list
        latex_string = "\n".join(latex_list) #convert list to string
        latex_string_list.append(latex_string) #add this csv's LaTeX table string to the full list of LaTeX table strings


    if tabular_string == "": #set default tabular format
        tabular_string = 'l' + 'c' * (number_of_columns - 1) #first column left aligned, rest centered

    #Set list of lines to be written to output file at beginning
    document_header_list = [r'\documentclass[' + str(font_size) + 'pt]{article}',r'\usepackage{amsmath}',r'\usepackage{pdflscape}',r'\usepackage[margin=0.3in]{geometry}',
                         r'\usepackage{dcolumn}',r'\usepackage{booktabs}',r'\usepackage{array}', r'\usepackage{threeparttable}',
                         r'\newcolumntype{L}[1]{>{\raggedright\let\newline\\\arraybackslash\hspace{0pt}}m{#1}}',
                         r'\newcolumntype{C}[1]{>{\centering\let\newline\\\arraybackslash\hspace{0pt}}m{#1}}',
                         r'\newcolumntype{R}[1]{>{\raggedleft\let\newline\\\arraybackslash\hspace{0pt}}m{#1}}',
                         r'\newcolumntype{.}{D{.}{.}{-1}}',r'\title{\LaTeX}',r'\date{}',r'\author{Nick DeRobertis}',
                         r'\begin{document}']

    table_header_list =  [r'\begin{table}',r'\centering',r'\begin{threeparttable}',
                         above_text,r'\caption{' + caption + '}',r'\begin{tabular}{' + tabular_string + '}',
                         r'\toprule']

    #Set list of lines to be written to output file at end
    table_footer_list = [r'\bottomrule',r'\end{tabular}',r'\begin{tablenotes}[para,flushleft]',r'\item ' + below_text,r'\end{tablenotes}',
                         r'\end{threeparttable}',r'\end{table}']

    document_footer_list = [r'\end{document}']

    if landscape:
        table_header_list.insert(0, r'\begin{landscape}')
        table_footer_list.append(r'\end{landscape}')

    if as_document:
        latex_header_list = document_header_list + table_header_list
        latex_footer_list = table_footer_list + document_footer_list
    else:
        latex_header_list = table_header_list
        latex_footer_list = table_footer_list

    #Actually write to file
    with open(outpath,'w') as f:
        for line in latex_header_list: #write each line in the header list, with carriage returns in between
            f.write(line)
            f.write("\n")
        for latex_string in latex_string_list: #write each csv table to file in LaTeX format
            f.write(latex_string)
        for line in latex_footer_list: #write each line in the footer list, with carriage returns in between
            f.write(line)
            f.write("\n")
        f.close()


    os.chdir(outfolder) #changes working filepath
    # Only create pdf if we are creating a standalone document
    if as_document:
        os.system('pdflatex ' + '"' + outname_tex + '"') #create PDF
    date_time_move_latex(outname, outfolder) #move table into appropriate date/number folder


from typing import Sequence, Union
from pyexlatex.models.item import SimpleItem
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.models.format.breaks import OutputLineBreak


class ContactLine(TextAreaMixin, SimpleItem):
    name = 'address'

    def __init__(self, contact_info: Union[str, Sequence[str]]):
        self.contact_info = contact_info
        self.add_data_from_content(contact_info)

        super().__init__(self.name, self._get_content())

    def _get_content(self):
        if isinstance(self.contact_info, str):
            return self.contact_info

        separator = f' {OutputLineBreak()} '
        return separator.join([str(item) for item in self.contact_info])



from typing import Optional, Any, Dict
import os

from jinja2 import FileSystemLoader

from pyexlatex.models.jinja import JinjaTemplate, JinjaEnvironment, DataString
from pyexlatex.models.template import Template


class Model(Template):
    """
    A data model combined with a Jinja template.

    Passes any class variables and methods to template to render the output string.
    Subclass this class to create your own data models for templates.

    Set exclude_attrs class variable as a list of strings of any attributes which should not be
    passed to the template.

    Examples:

        >>> import pyexlatex as pl
        >>> template_str = "{{ my_var | Bold }} {{ my_link }} {{ my_text }}"
        >>>
        >>> class MyModel(pl.Model):
        >>>     my_var = 'woo'
        >>>     my_link = pl.Hyperlink('https://www.google.com', 'Link!!!!')
        >>>
        >>>     def __init__(self, my_text: str, *args, **kwargs):
        >>>         self.my_text = my_text
        >>>         super().__init__(*args, **kwargs)
        >>>
        >>> model = MyModel('stuff', template_str=template_str)
        >>> str(model)
        '\\textbf{woo} \\href{https://www.google.com}{Link!!!!} stuff '



    """
    # Additional attributes to exclude adding to template variables
    exclude_attrs = []

    def __init__(self, template_str: Optional[str] = None, template_path: Optional[str] = None,
                 environment: Optional[JinjaEnvironment] = None):
        self.template_str = template_str
        self.template_path = template_path
        self._validate_template()
        self.environment = environment
        self.contents = self._get_contents()
        super().__init__()

    @property
    def environment(self):
        return self._environment

    @environment.setter
    def environment(self, env: Optional[JinjaEnvironment]):
        self._active_template_path = self.template_path
        if env is not None:
            self._environment = env
            return

        if self.template_path is not None:
            template_dir = os.path.dirname(os.path.realpath(self.template_path))
            template_name = os.path.basename(self.template_path)

            # Create environment with file system loader in the folder containing the template
            self._environment = JinjaEnvironment(loader=FileSystemLoader(template_dir))

            # Switch template path str passed to environment to just name, as environment was created in that folder
            self._active_template_path = template_name

        if self.template_str is not None:
            self._environment = JinjaEnvironment()


    @property
    def template(self) -> JinjaTemplate:
        if self.template_str is not None:
            return self.environment.from_string(self.template_str)

        # Get from path
        return self.environment.get_template(self._active_template_path)

    @property
    def render_dict(self) -> Dict[str, Any]:
        always_exclude_attrs = [
            'environment',
            'template',
            'template_str',
            'template_path',
            'render_dict',
            'contents',
            'format_contents',
            'add_data_from_content',
            'init_data',
            'data',
            'add_package',
            'equal_attrs',
            'exclude_attrs',
            'is_DocumentItem',
            'is_LatexItem',
            'join',
            'name',
            'next_level_down_class'
        ]
        full_exclude = always_exclude_attrs + self.exclude_attrs
        attrs = [item for item in dir(self) if item not in full_exclude and not item.startswith('_')]
        return {attr: getattr(self, attr) for attr in attrs}


    def _validate_template(self):
        if self.template_str is not None and self.template_path is not None:
            raise ValueError(f'only specify one of template_str, template_path. got {self.template_str} for '
                             f'template_str, {self.template_path} for template_path')
        if self.template_str is None and self.template_path is None:
            raise ValueError(f'must specify one of template_str, template_path. got {self.template_str} for '
                             f'template_str, {self.template_path} for template_path')

    def _get_contents(self) -> DataString:
        render_dict = self.render_dict
        self.add_data_from_content(render_dict)
        return self.template.render(**render_dict)



from pyexlatex.table.models.panels import Panel
from pyexlatex.table.models.data.table import DataTable
from pyexlatex.table.models.labels.table import LabelCollection, LabelTable
from pyexlatex.table.models.labels.label import Label
from pyexlatex.table.models.table.table import Table
from pyexlatex.table.models.texgen.items import Tabular
from pyexlatex.table.models.data.valuestable import ValuesTable
from pyexlatex.table.models.texgen.alignment import ColumnAlignment, ColumnsAlignment
from pyexlatex.table.models.texgen.lines import TopRule, MidRule, BottomRule, TableLineSegment
from pyexlatex.table.models.labels.multicolumn import MultiColumn
from pyexlatex.models.format.breaks import TableLineBreak
from pyexlatex.table.models.texgen.tabularstar import TabularStar



from typing import Optional, Union

from mixins.repr import ReprMixin

from pyexlatex.models import Item
from pyexlatex.models.containeritem import ContainerItem
from pyexlatex.models.sizes.textwidth import TextWidth
from pyexlatex.table import ColumnsAlignment
from pyexlatex.table.models.texgen.tabularbase import BaseTabular


class TabularStar(ContainerItem, Item, ReprMixin, BaseTabular):
    """
    tabular* environment, allows for controlling width. Gives additional column alignment X to fill available area
    """
    name = 'tabular*'
    repr_cols = ['align']

    def __init__(self, content, align: Optional[Union[ColumnsAlignment, str]] = None, width: Optional = None,
                 fill_between: bool = True):
        if not isinstance(content, (list, tuple)):
            content = [content]
        self.align = self._get_columns_alignment_from_passed_align(content, align)

        if width is None:
            width = TextWidth()
        self.width = width
        self.fill_between = fill_between

        # TODO: really this should be cmidrule and others that require booktabs, but need to get all nested table
        # TODO: structure aggregating data.
        self.add_package('booktabs')
        self.add_data_from_content(self.align)
        super().__init__(self.name, content, env_modifiers=self._modifier_str)

    @property
    def _modifier_str(self) -> str:
        return self._wrap_with_braces(self.width) + self._align_str

    @property
    def _align_str(self) -> str:
        base_str = str(self.align)
        if self.fill_between:
            # TODO: make models for this
            base_str = r'@{\extracolsep{\fill}}' + base_str
        return self._wrap_with_braces(base_str)


from typing import Sequence, Optional

from pyexlatex.models.item import NoOptionsNoContentsItem


class TableOfContents(NoOptionsNoContentsItem):
    name = 'tableofcontents'

    def __init__(self, options: Optional[Sequence[str]] = ('currentsection',), **kwargs):
        self.options = options
        super().__init__(self.name, modifiers=self.options_str, **kwargs)

    @property
    def options_str(self) -> str:
        options_str = ', '.join(self.options)
        return self._wrap_with_bracket(options_str)



from typing import Optional
from copy import deepcopy
from pyexlatex.models import Item
from pyexlatex.models.section.base import TextAreaBase
from pyexlatex.presentation.beamer.frame.title import FrameTitle
from pyexlatex.constants.flags import HAS_MINTED


class Frame(TextAreaBase):
    """
    Basic unit for creating a slide in a presentation.
    """
    name = 'frame'

    def __init__(self, content, title: Optional[str] = None, label: Optional[str] = None, **kwargs):
        self.title = title
        if isinstance(content, (Item, str)):
            content = [content]
        else:
            content = deepcopy(content)  # don't modify content inplace

        if self.title is not None:
            content.insert(0, FrameTitle(title))

        self.add_data_from_content(content)

        self.content = content
        super().__init__(self.name, self.content, label=label, env_modifiers=self._get_env_modifiers(), **kwargs)

    @property
    def is_fragile(self) -> bool:
        if HAS_MINTED in self.data.flags:
            # Minted package requires using fragile frames
            return True

        return False

    def _get_env_modifiers(self) -> str:
        modifiers = []
        if self.is_fragile:
            modifiers.append('fragile')
        # Add any other fragile conditions here

        if not modifiers:
            return ''

        modifiers = ', '.join(modifiers)
        return self._wrap_with_bracket(modifiers)


from typing import Optional, Sequence, List, Tuple, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay import Overlay
    from pyexlatex.graphics.tikz import Node
    from pyexlatex.graphics.shape import Shape
from pyexlatex.graphics.tikz.item import TikZItem

PATH_TYPES = (
    'draw',
    'fill',
    'pattern',
    'shade',
    'clip',
    'use as bounding box'
)

Location = Union[Tuple[float, float], 'Node']


class Path(TikZItem):
    """
    Lower-level class for drawing individual lines or shapes, should only be needed to be used for very custom graphics
    """
    name = 'path'


    def __init__(self, path_type: str, points: Sequence[Union[Tuple[float, float], 'Node']],
                 draw_type: str = '--',
                 options: Optional[Sequence[str]] = None,
                 overlay: Optional['Overlay'] = None):
        self.points = points
        self.draw_type = draw_type
        self.path_type = path_type.lower()
        self.options = options
        self._validate()

        super().__init__(self.name, self.path_str, options=self.all_options, overlay=overlay)


    def _validate(self):
        self._validate_path_type()

    def _validate_path_type(self):
        if self.path_type not in PATH_TYPES:
            raise ValueError(f'invalid path_type, got {self.path_type}, expected one of {", ".join(PATH_TYPES)}')

    @property
    def all_options(self) -> List[str]:
        options = [self.path_type]
        if self.options is not None:
            options.extend(self.options)
        return options

    @property
    def draw_join_str(self) -> str:
        return f' {self.draw_type} '

    @property
    def path_str(self) -> str:
        locations: List[str] = []
        for location in self.points:
            if hasattr(location, 'is_Node') and location.is_Node:
                location: 'Node'
                locations.append(f'({location.label})')
            elif hasattr(location, 'shape_node'):
                location: 'Shape'
                locations.append(f'({location.shape_node.label})')
            else:
                location: Tuple[float, float]
                locations.append(str(location))
        if len(locations) == 1:
            # Got a single point draw path, such as circle
            return f'{locations[0]} {self.draw_type}'
        return self.draw_join_str.join(locations)


class SpecificPath(Path):
    """
    Subclass this path to build specific paths such as rectangle, circle, arrow
    """
    draw_type = '<invalid, do not use SpecificPath directly>'
    path_type = '<invalid, do not use SpecificPath directly>'

    def __init__(self, points: Sequence[Tuple[float, float]], *args, **kwargs):
        super().__init__(self.path_type, points, *args, draw_type=self.draw_type, **kwargs)



from typing import List, Optional, Dict, Union, Sequence, Callable
from copy import deepcopy

from pyexlatex.models.environment import Environment
from pyexlatex.models.item import Item, ItemBase
from pyexlatex.models.control.documentclass.documentclass import DocumentClass
from pyexlatex.models.package import Package
from pyexlatex.models.page.style import PageStyle
from pyexlatex.models.landscape import Landscape
from pyexlatex.logic.pdf.main import document_to_pdf_and_move, latex_str_to_pdf_obj_with_sources
from pyexlatex.texgen.replacements.filename import latex_filename_replacements
from pyexlatex.logic.extract.docitems import extract_document_items_from_ambiguous_collection
from pyexlatex.models.page.number import right_aligned_page_numbers
from pyexlatex.models.page.header import remove_header
from pyexlatex.models.page.footer import CenterFooter
from pyexlatex.models.format.sectionnum import SectionNumberingFormatter
from pyexlatex.typing import AnyItem, ItemOrListOfItems, ItemAndPreEnvContents
from pyexlatex.models.commands.endfloat import DeclareDelayedFloatFlavor
from pyexlatex.models.format.text.linespacing import LineSpacing
from pyexlatex.models.commands.floatrow import DeclareFloatFont, FloatSetup
from pyexlatex.models.containeritem import ContainerItem
from pyexlatex.models.page.header import Header
from pyexlatex.models.control.setcounter import SetCounter


class DocumentEnvironment(Environment):
    name = 'document'

    def __init__(self):
        super().__init__(name=self.name)


class DocumentBase(ContainerItem, Item):
    name = '<invalid, do not use DocumentBase directly>'
    document_class_obj = None

    def __init__(self, content: ItemOrListOfItems, packages: List[Package]=None,
                 pre_env_contents: Optional[ItemOrListOfItems] = None, data_cleanup_func: Optional[Callable] = None,
                 pre_output_func: Optional[Callable] = None):
        """

        :param content:
        :param packages:
        :param pre_env_contents:
        :param data_cleanup_func: should accept DocumentSetupData and modify it in place. This is called just before
            using the data.
        :param pre_output_func: function which modifies the latex string before outputting it. The function should
            accepts a single argument which is a string of the entire latex contents, and it should return a string
            which will be used as the latex contents for output.
        """
        from pyexlatex.logic.builder import build, _build
        self.add_data_from_content(content)
        self.add_data_from_content(self.document_class_obj)

        if packages is not None:
            self.data.packages.extend(packages)

        self.has_references = False
        if self.data.references:
            self.has_references = True

        if isinstance(content, (ItemBase, str)):
            content = [content]

        if pre_env_contents is None:
            pre_env_contents = []

        if data_cleanup_func:
            data_cleanup_func(self.data)

        possible_pre_env_contents = [
            self.document_class_obj,
            *self.data.begin_document_items,
            *[str(package) for package in self.data.packages],
            *pre_env_contents
        ]

        self.pre_env_contents = build([item for item in possible_pre_env_contents if item is not None])

        content.extend(self.data.end_document_items)

        self.contents = content


        content = deepcopy(content)  # don't overwrite original objects
        # combine content into a single str
        content = build(content)

        if pre_output_func:
            content = pre_output_func(content)

        super().__init__(self.name, content, pre_env_contents=self.pre_env_contents)

    def _repr_pdf_(self):
        tex = str(self)

        return latex_str_to_pdf_obj_with_sources(
            tex,
            image_paths=self.data.filepaths,
            image_binaries=self.data.binaries,
            run_bibtex=self.has_references
        ).readb()

    def to_pdf(self, outfolder, outname='document',
               move_folder_name='Tables', as_document=True,
               date_time_move: bool = False):
        tex = str(self)

        outname = latex_filename_replacements(outname)

        document_to_pdf_and_move(
            tex,
            outfolder=outfolder,
            outname=outname,
            image_paths=self.data.filepaths,
            move_folder_name=move_folder_name,
            as_document=as_document,
            image_binaries=self.data.binaries,
            run_bibtex=self.has_references,
            date_time_move=date_time_move
        )

    @classmethod
    def from_ambiguous_collection(cls, collection, **document_kwargs):
        content = extract_document_items_from_ambiguous_collection(collection)

        return cls(content, **document_kwargs)


class Document(DocumentBase):
    """
    Main class used for creating latex documents.
    """
    name = 'document'

    def __init__(self, content: ItemOrListOfItems, packages: List[Package]=None, landscape=False,
                 title: str=None, authors: Optional[Union[List[str], str]] = None, date: str=None, abstract: str=None,
                 skip_title_page: bool=False,
                 page_modifier_str: Optional[str]='margin=0.8in, bottom=1.2in', page_header: bool=False,
                 page_numbers: bool=True, appendix_modifier_str: Optional[str] = 'page',
                 section_numbering_styles: Optional[Dict[str, str]] = None, floats_at_end: bool = False,
                 floats_at_end_options: str = 'nolists',
                 document_type: str = 'article', font_size: Optional[float] = None,
                 num_columns: Optional[int] = None, line_spacing: Optional[float] = None,
                 tables_relative_font_size: int = 0, figures_relative_font_size: int = 0,
                 page_style: str = 'fancy', custom_headers: Optional[Sequence[Header]] = None,
                 remove_section_numbering: bool = False, separate_abstract_page: bool = False,
                 extra_title_page_lines: Optional[Sequence] = None, empty_title_page_style: bool = False,
                 pre_output_func: Optional[Callable] = None):
        from pyexlatex.models.title.page import TitlePage

        all_packages = self.construct_packages(
            packages=packages,
            page_modifier_str=page_modifier_str,
            appendix_modifier_str=appendix_modifier_str,
            floats_at_end=floats_at_end,
            floats_at_end_options=floats_at_end_options,
            line_spacing=line_spacing,
            tables_relative_font_size=tables_relative_font_size,
            figures_relative_font_size=figures_relative_font_size
        )

        if section_numbering_styles is None:
            section_numbering_styles = {}

        section_num_styles = SectionNumberingFormatter.list_from_string_format_dict(section_numbering_styles)

        if isinstance(content, (Item, str)):
            content = [content]
        if authors is None:
            authors = []
        if custom_headers is None:
            custom_headers = []
        if isinstance(authors, (Item, str)):
            authors = [authors]

        self.document_class_obj = DocumentClass(
            document_type=document_type,
            font_size=font_size,
            num_columns=num_columns
        )

        possible_extra_pre_env_contents = [
            *section_num_styles,
            SetCounter('secnumdepth', 0) if remove_section_numbering else None,
            PageStyle(page_style),

            # header is there by default. add remove header lines if page_header=False
            remove_header if not page_header and not custom_headers else None,
            *custom_headers,

            # add right page numbers. if not, use blank center footer to clear default page numbers in center footer
            right_aligned_page_numbers if page_numbers else CenterFooter('')
        ]

        pre_env_contents = [item for item in possible_extra_pre_env_contents if item is not None]

        if not skip_title_page and _should_create_title_page(title=title, authors=authors, date=date, abstract=abstract):
            title_page = TitlePage(
                title=title,
                authors=authors,
                date=date,
                abstract=abstract,
                separate_abstract_page=separate_abstract_page,
                extra_lines=extra_title_page_lines,
                empty_page_style=empty_title_page_style
            )
            content.insert(0, title_page)
            self.has_title_page = True
        else:
            self.has_title_page = False

        if landscape:
            content = Landscape().wrap(content)

        super().__init__(
            content,
            packages=all_packages,
            pre_env_contents=pre_env_contents,
            pre_output_func=pre_output_func
        )

    def __repr__(self):
        return f'<Document>'

    def construct_packages(self, packages: List[Package]=None,
                           page_modifier_str: Optional[str]='margin=0.8in, bottom=1.2in',
                           appendix_modifier_str: Optional[str] = 'page',
                           floats_at_end: bool = False, floats_at_end_options: str = 'nolists',
                           line_spacing: Optional[float] = None,
                           tables_relative_font_size: int = 0, figures_relative_font_size: int = 0) -> List[Package]:
        from pyexlatex.texgen.packages.default import default_packages

        if packages is None:
            packages = default_packages.copy()

        if page_modifier_str is not None:
            # Set margins, body size, etc. with geometry package
            packages.append(Package('geometry', modifier_str=page_modifier_str))

        if tables_relative_font_size or figures_relative_font_size:
            packages.append(Package('floatrow'))
            if tables_relative_font_size:
                declared_font = DeclareFloatFont(tables_relative_font_size)
                float_setup_str = f'font={declared_font.size_def.name},capposition=top'
                packages.extend([
                    declared_font,
                    FloatSetup('table', float_setup_str),
                    # FloatSetup('ltable', float_setup_str)
                ])
            if figures_relative_font_size:
                declared_font = DeclareFloatFont(figures_relative_font_size)
                float_setup_str = f'font={declared_font.size_def.name},capposition=top'
                packages.extend([
                    declared_font,
                    FloatSetup('figure', float_setup_str),
                    # FloatSetup('lfigure', float_setup_str)
                ])

        if floats_at_end:
            packages.extend([
                Package('endfloat', modifier_str=floats_at_end_options if floats_at_end_options else None),
                DeclareDelayedFloatFlavor('ltable', 'table'),  # treat custom environment ltable (landscape table) as table
                DeclareDelayedFloatFlavor('lfigure', 'figure') # treat custom environment lfigure (landscape figure) as figure
            ])

        if line_spacing:
            packages.extend([
                Package('setspace'),
                LineSpacing(line_spacing)
            ])

        packages.append(Package('appendix', modifier_str=appendix_modifier_str))

        return packages


def _should_create_title_page(title: str = None, authors: Optional[List[str]] = None, date: str = None,
                              abstract: str = None):
    return any([
        title is not None,
        authors is not None and authors != [],
        date is not None,
        abstract is not None
    ])

def _content_items_and_collected_pre_env_contents():
    raise NotImplementedError('see TODO in _standardize_content_item_for_inclusion_in_document')

def _standardize_content_item_for_inclusion_in_document(item: AnyItem) -> ItemAndPreEnvContents:

    # No extra processing needed if not Document
    if not isinstance(item, Document):
        return item, None

    # TODO: restructure to remove title pages
    # TODO: restructure to extract content from Document

    return item, item.pre_env_contents




from typing import List


class LatexError:

    def __init__(self, filename: str, line: int, context: List[str]):
        self.filename = filename
        self.line = line
        self.context = context

    @property
    def error(self):
        strip_chars = len(f'{self.filename}:{self.line}: ')
        return self.text[strip_chars:]

    @property
    def text(self):
        return ''.join(self.context)

    def __repr__(self):
        return f'<LatexError({self.error})>'



from pyexlatex.table.models.panels.collection import PanelCollection
from pyexlatex.table.models.texgen.items import Table, TableDocument
from pyexlatex.logic.pdf.main import document_to_pdf_and_move


def output_panel_collection_to_tex(panel_collection: PanelCollection, out_folder: str, out_name: str='table',
                                   as_document=True,
                                   tabular_kwargs={}, three_part_table_kwargs={},
                                   table_kwargs={}, document_kwargs={}):
    if as_document:
        table_or_document = TableDocument.from_panel_collection(
            panel_collection,
            tabular_kwargs=tabular_kwargs,
            three_part_table_kwargs=three_part_table_kwargs,
            table_kwargs=table_kwargs,
            **document_kwargs
        )
    else:
        table_or_document = Table.from_panel_collection(
            panel_collection,
            tabular_kwargs=tabular_kwargs,
            three_part_table_kwargs=three_part_table_kwargs,
            **table_kwargs
        )


    document_to_pdf_and_move(
        table_or_document,
        outfolder=out_folder,
        outname=out_name,
        as_document=as_document,
        move_folder_name='Tables'
    )


from pyexlatex.graphics.tikz.node.position.directions import Above, Right, Left, Below
from pyexlatex.graphics.tikz.path import Path
from pyexlatex.graphics.tikz.picture import TikZPicture
from pyexlatex.graphics.tikz.node.node import Node
from pyexlatex.graphics.arrow import Arrow
from pyexlatex.graphics.flowcharts.linear import LinearFlowchart
from pyexlatex.graphics.rectangle import Rectangle
from pyexlatex.graphics.modified_picture import ModifiedPicture



from pyexlatex.models.format.code.minted import Minted


class Python(Minted):
    """
    Renders passed contents as Python code, with syntax highlighting.
    """

    def __init__(self, contents):
        super().__init__(contents, 'python')



from typing import Optional, Sequence
from pyexlatex.models.item import Item
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.models.package import Package


class TikZPicture(TextAreaMixin, Item):
    """
    Container which holds any graphics.
    """
    name = 'tikzpicture'

    def __init__(self, contents, env_modifiers: Optional[Sequence[str]]=None):
        self.init_data()
        self.data.packages.append(Package('tikz'))
        super().__init__(self.name, contents, env_modifiers=env_modifiers)




import shutil
import os
import time
from typing import Iterable, Callable, Optional


def move_if_exists(inpath: str, outpath: str, retries: int = 3, sleep_time: float = 0.1) -> None:
    _run_func_handling_file_modify_exceptions(
        shutil.move,
        inpath,
        outpath,
        func_short_desc='move',
        retries_remaining=retries,
        sleep_time=sleep_time
    )


def remove_all_if_exist(filepaths: Iterable[str]):
    [remove_if_exists(filepath) for filepath in filepaths]


def remove_if_exists(filepath: str, retries: int = 3, sleep_time: float = 0.1):
    _run_func_handling_file_modify_exceptions(
        os.remove,
        filepath,
        func_short_desc='delete',
        retries_remaining=retries,
        sleep_time=sleep_time
    )


def move_all_if_exists(inpaths: Iterable[str], outfolder: str):
    [move_if_exists(inpath, outfolder) for inpath in inpaths]


def _run_func_handling_file_modify_exceptions(func: Callable, filepath: str, *args,
                                              func_short_desc: Optional[str] = None, retries_remaining: int = 3,
                                              sleep_time: float = 0.1,
                                              **kwargs) -> None:
    if retries_remaining < 0:
        raise PermissionError(f'could not modify {filepath}, even after retrying')

    func_desc = _get_func_short_description(func, func_short_desc)

    try:
        func(filepath, *args, **kwargs)
    except FileNotFoundError:
        print(f'Cannot {func_desc}: did not find {filepath}')
        return
    except PermissionError:
        # Retry with waits before determining that the file cannot be modified
        time.sleep(sleep_time)
        return _run_func_handling_file_modify_exceptions(
            func,
            filepath,
            *args,
            func_short_desc=func_short_desc,
            retries_remaining=retries_remaining - 1,
            sleep_time=sleep_time
        )



def _get_func_short_description(func: Callable, func_short_desc: Optional[str] = None) -> str:
    if func_short_desc is None:
        return f'call {func.__name__}'
    else:
        return func_short_desc


from pyexlatex.models.item import SimpleItem
from pyexlatex.models.section.base import TextAreaMixin


class AtBeginSection(TextAreaMixin, SimpleItem):
    name = 'AtBeginSection'

    def __init__(self, contents, **kwargs):
        super().__init__(self.name, contents, **kwargs)



from typing import Optional
from pyexlatex.models.item import SimpleItem
from pyexlatex.models.package import Package
from pyexlatex.models.control.filecontents import FileContents


class BibliographyStyle(SimpleItem):
    name = 'bibliographystyle'

    def __init__(self, contents):
        super().__init__(self.name, contents)


class Style(BibliographyStyle):

    def __init__(self, style_name: str, style_definition: Optional[str] = None):
        self.style_name = style_name
        self.style_definition = style_definition
        
        self.init_data()
        if style_definition is not None:
            self.data.filepaths.append(self.style_filename)
            self.data.binaries.append(bytes(self.style_definition, 'utf8'))

        super().__init__(self.style_source_name)


    @property
    def style_filename(self) -> str:
        return self.style_name + '.bst'

    @property
    def style_source_name(self) -> str:
        return 'Sources/' + self.style_name


from typing import Union
from pyexlatex.models.template import Template
from pyexlatex.models.format.fills import VFill, HFill


class SpacedBase(Template):
    """
    Base class for VerticallySpaced and HorizontallySpaced. Handles the logic of adding VFill and HFill between and
    around items.
    """

    def __init__(self, content, vertically_space: bool, pad_ends: bool):
        self.orig_content = content
        self.vertically_space = vertically_space
        self.pad_ends = pad_ends
        self.contents = self._get_contents()
        super().__init__()

    def _get_contents(self):
        if not isinstance(self.orig_content, (list, tuple)):
            orig_content = [self.orig_content]
        else:
            orig_content = self.orig_content

        orig_valid_content = [item for item in orig_content if item is not None]

        contents = []
        for content in orig_valid_content:
            contents.extend([
                content,
                self.spacer
            ])
        del contents[-1]  # strip final spacer

        if self.pad_ends:
            # Add additional spacers at beginning and end
            contents.insert(0, self.spacer)
            contents.append(self.spacer)

        return contents


    @property
    def spacer(self) -> Union[VFill, HFill]:
        if self.vertically_space:
            return VFill()
        return HFill()


class VerticallySpaced(SpacedBase):

    def __init__(self, content, pad_ends: bool = False):
        super().__init__(content, vertically_space=True, pad_ends=pad_ends)


class HorizontallySpaced(SpacedBase):

    def __init__(self, content, pad_ends: bool = False):
        super().__init__(content, vertically_space=False, pad_ends=pad_ends)



from pyexlatex.models.item import SimpleItem


class RightFooter(SimpleItem):
    name = 'rfoot'

    def __init__(self, contents):
        super().__init__(self.name, contents)

class CenterFooter(SimpleItem):
    name = 'cfoot'

    def __init__(self, contents):
        super().__init__(self.name, contents)


from typing import Sequence


def item_is_in_allowed_type_strs(item, allowed_values: Sequence[str]) -> bool:
    is_allowed_attrs = [f'is_{value}' for value in allowed_values]
    for allowed_attr in is_allowed_attrs:
        if hasattr(item, allowed_attr) and getattr(item, allowed_attr):
            return True

    return False


from string import ascii_uppercase

def panel_string(i: int):
    return f'Panel {_panel_character(i)}: '

def _panel_character(i: int):
    if i < 0:
        raise ValueError('must pass positive index for panel character')
    if i > 26:
        raise NotImplementedError('need to work on panel letters greater than Z')

    return ascii_uppercase[i]



from typing import Sequence
from pyexlatex.models.section.base import TextAreaBase
from pyexlatex.models.item import Item


class ParagraphIndent(TextAreaBase, Item):
    """
    Used for setting paragraph indents, from the left or right.
    """
    name = 'adjustwidth'

    def __init__(self, contents, left_adjust: int = 1, right_adjust: int = 0):
        self.left_adjust = left_adjust
        self.right_adjust = right_adjust
        self.init_data()
        self.data.packages.append('changepage')
        super().__init__(self.name, contents, env_modifiers=self.options_str)

    @property
    def options_str(self) -> str:
        options_str = ''
        if self.left_adjust > 0:
            options_str += self._wrap_with_braces(f'{self.left_adjust}cm')
        else:
            options_str += '{}'
        if self.right_adjust > 0:
            options_str += self._wrap_with_braces(f'{self.right_adjust}cm')
        else:
            options_str += '{}'
        return options_str



from typing import Optional
from pyexlatex.models.format.text.color.deftypes.base import ColorDefinition


class Hex(ColorDefinition):
    definition_type = 'HTML'

    def __init__(self, hex_code: str, color_name: Optional[str] = None):
        self.hex_code = hex_code.strip('#')
        super().__init__(color_content=self.hex_code, color_name=color_name)



from pyexlatex import Template
from pyexlatex.models.control.newcounter import NewCounter
from pyexlatex.models.control.setcounter import SetCounter
from pyexlatex.models.control.value import Value
from pyexlatex.models.item import NoOptionsNoContentsItem


class _Appendix(NoOptionsNoContentsItem):
    name = 'appendix'

    def __init__(self, **kwargs):
        super().__init__(self.name, **kwargs)


class Appendix(Template):

    def __init__(self, content):
        if not isinstance(content, (list, tuple)):
            content = [content]
        self.content = content
        self.contents = self._get_contents()
        super().__init__()

    def _get_contents(self):
        pre_content_contents = [
            NewCounter('finalframe'),
            SetCounter('finalframe', Value('framenumber'))
        ]

        post_content_contents = [
            SetCounter('framenumber', Value('finalframe'))
        ]

        return [
            _Appendix(),
            *pre_content_contents,
            *self.content,
            *post_content_contents
        ]


from pyexlatex.models.item import SimpleItem


class Monospace(SimpleItem):
    """
    Pass text to render it in a monospace font.
    """
    name = 'texttt'

    def __init__(self, content, **kwargs):
        super().__init__(self.name, content, **kwargs)


from typing import Optional
from pyexlatex.models.item import SimpleItem


class FrameTitle(SimpleItem):
    name = 'frametitle'

    def __init__(self, title: str, modifiers: Optional[str] = None):
        self.title = title
        super().__init__(self.name, title, modifiers)


from typing import List, Optional, Sequence
from pyexlatex.models.template import Template
from pyexlatex.models.title.title import Title
from pyexlatex.models.credits.author import Author
from pyexlatex.models.date import Date
from pyexlatex.models.section.abstract import Abstract
from pyexlatex.texgen import _maketitle_str
from pyexlatex.logic.format.and_join import join_with_commas_and_and
from pyexlatex.models.sizes.baselineskip import BaselineSkip
from pyexlatex.models.control.group import Group
from pyexlatex.models.control.setlength import SetLength
from pyexlatex.models.control.setcounter import SetCounter
from pyexlatex.models.format.text.bold import Bold
from pyexlatex.models.sizes.textsizes import TextSize
from pyexlatex.models.format.centering import Center
from pyexlatex.models.control.newpage import PageBreak
from pyexlatex.models.page.style import ThisPageStyle
from pyexlatex.models.format.vspace import VSpace
from pyexlatex.models.format.hspace import HSpace
from pyexlatex.models.format.breaks import LineBreak


class TitlePage(Template):

    def __init__(self, title: str = None, authors: List[str] = None, date: Optional[str] = None,
                 abstract: Optional[str] = None, separate_abstract_page: bool = False,
                 extra_lines: Optional[Sequence] = None, empty_page_style: bool = False):

        self.empty_page_style = empty_page_style
        self.title = title

        if authors is not None:
            author = join_with_commas_and_and(authors)
        else:
            author = None

        contents = [
            self._title_obj,
            Author(author, short_author=None) if author is not None else None,
            Date(date) if date is not None else Date(),
            _maketitle_str() if title is not None else None,
            ThisPageStyle('empty') if empty_page_style else None,
            Abstract(abstract) if abstract is not None and not separate_abstract_page else None,
        ]

        if extra_lines is not None and not separate_abstract_page:
            contents.extend(extra_lines)

        if separate_abstract_page and abstract:
            contents.append(
                CustomTitlePage(title, abstract, extra_lines=extra_lines, next_page_number=1)
            )

        self.contents = [content for content in contents if content is not None]

        super().__init__()

    @property
    def _title_obj(self) -> Optional[Title]:
        if self.title is None:
            return None

        title = self.title

        if self.empty_page_style:
            # Empty page style strips away bold title, bring it back manually
            title = Bold(title)

        return Title(title)


class CustomTitlePage(Template):
    r"""
    Title page which is not created using \maketitle, instead it is created manually and so may be included at
    any point in a document, and there may be multiple.
    """

    def __init__(self, title: str, abstract: Optional[str] = None, title_relative_size: int = 3,
                 gap_after_title_pt: int = 35, extra_lines: Optional[Sequence] = None,
                 next_page_number: Optional[int] = None):
        self.title = title
        self.abstract = abstract
        self.title_relative_size = title_relative_size
        self.gap_after_title = gap_after_title_pt
        self.extra_lines = extra_lines
        self.next_page_number = next_page_number

        self.contents = self._get_contents()
        super().__init__()

    def _get_contents(self):
        contents = [
            PageBreak(),
            ThisPageStyle('empty'),
            self._title_block,
            VSpace(self.gap_after_title, units='pt'),
        ]
        if self.abstract is not None:
            contents.extend(self._abstract_block)
        if self.extra_lines is not None:
            contents.extend(self._extra_lines_block)
        contents.append(PageBreak())
        if self.next_page_number is not None:
            contents.append(SetCounter('page', 1))
        return contents

    @property
    def _title_block(self) -> Group:
        main_contents = [
            SetLength(BaselineSkip(), '40pt'),
            TextSize(self.title_relative_size),
            Bold(self.title)
        ]
        return Group(Center(main_contents))

    @property
    def _abstract_block(self):
        if self.abstract is None:
            return None
        abstract_title = Center(Bold('Abstract'))
        return [
            abstract_title,
            VSpace(-10, 'pt'),
            HSpace(1.5, 'em'),
            self.abstract
        ]

    @property
    def _extra_lines_block(self):
        if self.extra_lines is None:
            return None
        contents = [VSpace(20, 'pt')]
        for line in self.extra_lines:
            contents.extend([
                line,
                LineBreak(),
            ])
        return contents









from pyexlatex.models.item import SimpleItem


class Ref(SimpleItem):
    """
    Pass a label to create an in text reference to that label. Will display the number of the section in the text.
    """
    name = 'ref'

    def __init__(self, contents):
        super().__init__(self.name, contents)


class NameRef(SimpleItem):
    """
    Pass a label to create an in text reference to that label. Will display the name of the section in the text.
    """
    name = 'nameref'

    def __init__(self, contents):
        self.add_package('nameref')
        super().__init__(self.name, contents)



from typing import Sequence, Any, List, Tuple
from pyexlatex.presentation.beamer.columns.column import Column
from pyexlatex.presentation.beamer.columns.columns import Columns
from pyexlatex.layouts.base import LayoutBase


class CellLayout(LayoutBase):
    """
    Lay content out in a grid where each cell is arbitrarily sized
    """

    def __init__(self, content: Sequence[Sequence[Any]], grid_shape: Sequence[Sequence[Tuple[float, float]]], **kwargs):
        """

        Args:
            content:
            grid_shape: list of list of tuples of (width, height), e.g. [
                    [(0.3, 0.4), (0.7, 0.4)],
                    [(0.3, 0.6), (0.7, 0.6)],
                ] would create a 2x2 grid with 40% of height in the first row, 60% in the second row. 30% width in the
                first column, 70% width in the second column.
            **kwargs:
        """
        self.grid_shape = grid_shape
        super().__init__(content, **kwargs)
        self._validate()
        self.content = self._get_layout()

    def _validate(self):
        value_error = ValueError(f'must pass content and grid_shape of the same dimensions, '
                                 f'got content {self.contents} and grid_shape {self.grid_shape}')
        if len(self.contents) != len(self.grid_shape):
            raise value_error

        for i, content_row in enumerate(self.contents):
            shape_row = self.grid_shape[i]
            if len(shape_row) != len(content_row):
                raise value_error

    def _get_layout(self) -> List[Columns]:
        content_in_columns = []
        for row_num, shape_row in enumerate(self.grid_shape):
            cells = []
            for col_num, (width, height) in enumerate(shape_row):
                cell_content = self.contents[row_num][col_num]
                cell = Column(
                    cell_content,
                    frac_of_text_width=self._rescale_width(width, len(shape_row)),
                    frac_of_page_height=self._rescale_height(height)
                )
                cells.append(cell)
            row = Columns(cells)
            content_in_columns.append(row)

        return content_in_columns

    @staticmethod
    def _rescale_height(height: float):
        """
        With the standard Madrid header and footer, 0.8\textheight is about the full area with a bit of padding on
        top and bottom. Therefore passing 0.5 should be rescaled to 50% of the available area, or 0.4\textheight
        """
        # TODO: think about other themes having different text area
        available_area = 0.8
        return height * available_area

    @staticmethod
    def _rescale_width(width: float, num_cols: int):
        """
        Latex adds padding inbetween the columns, so 100% area is not really available. Need to set the available
        area based on the number of columns
        Args:
            width:
            num_cols:

        Returns:
        """
        # TODO: figure out how available area changes based on number of columns. Defaulting to doing nothing now
        available_area = 1
        return width * available_area


class GridLayout(CellLayout):
    """
    Evenly spaced grid layout.

    Creates a CellLayout, automatically setting widths and heights as an even split based on the shape
    of the content passed to make a grid
    """

    def __init__(self, content: Sequence[Sequence[Any]], **kwargs):
        grid_shape = []
        num_rows = len(content)
        height = 1/num_rows
        for row in content:
            num_columns = len(row)
            width = 1/num_columns
            grid_shape.append([(width, height) for _ in range(len(row))])
        super().__init__(content, grid_shape, **kwargs)


from typing import Union, Iterable

from pyexlatex.table.models.data.dataitem import DataItem
from pyexlatex.table.models.labels.collection import LabelCollection
from pyexlatex.table.models.labels.label import Label
from pyexlatex.table.models.spacing.cell import CellSpacer
from pyexlatex.table.models.table.rowbase import RowBase
from pyexlatex.texparser.clean import _remove_backslashes


class Row(RowBase):
    repr_cols = ['values']

    def __init__(self, values: Union[Iterable[DataItem], LabelCollection, Label]):

        # Don't allow nested rows. If the only values passed to a Row are a Row, then use the values of that
        # row rather than the row itself as values
        if isinstance(values, Row):
            self.values = values.values
        else:
            self.values = values

    def __str__(self):
        str_list = []
        for value in self.values:
            # Handle other forms of blanks
            if isinstance(value, Iterable) and len(value) == 1 and str(value[0]).strip() == '':
                str_list.append(' ')
            # The usual case, just join string representation of items
            else:
                str_list.append(str(value))

        return ' & '.join(str_list)

    @classmethod
    def from_latex_row_str(cls, latex_row_str):
        # Split before removing backslashes so that \& doesn't get counted as &
        str_values = latex_row_str.split(' & ')
        str_values = [_remove_backslashes(val) for val in str_values]
        values = [DataItem(value) for value in str_values]
        return cls(values)


from pyexlatex.layouts.grid import CellLayout, GridLayout
from pyexlatex.layouts.multicol import MultiCol
from pyexlatex.layouts.spaced import VerticallySpaced, HorizontallySpaced
from pyexlatex.models.format.paragraph.multicol import MultiColumn



from typing import Union, AnyStr, List

import numpy as np
import pandas as pd

from mixins.repr import ReprMixin
from pyexlatex.table.logic.panels.combine import (
    common_column_labels,
    common_row_labels,
    remove_label_collections_from_grid
)
from pyexlatex.table.logic.panels.letters import panel_string
from pyexlatex.table.logic.panels.topleft import _set_top_left_corner_labels
from pyexlatex.table.models.labels.table import LabelTable, LabelCollection
from pyexlatex.table.models.panels.panel import Panel
from pyexlatex.table.models.panels.panel import PanelGrid, GridShape
from pyexlatex.table.models.spacing.columntable import ColumnPadTable
from pyexlatex.table.models.spacing.rowtable import RowPadTable
from pyexlatex.table.models.table.section import TableSection


class PanelCollection(ReprMixin):
    """
    Lays panel contents in a grid, consolidates labels, creates padding between tables

    """
    repr_cols = ['name', 'panels']

    def __init__(self, panels: [Panel], label_consolidation: str='object', enforce_label_order=True,
                 top_left_corner_labels: Union[LabelTable, LabelCollection, List[AnyStr], AnyStr]=None,
                 pad_rows: int=1, pad_columns: int=1, name: str=None):
        """

        :param panels: list of Panels, which represent a full set of rows of the table. for multiple
                        sections in one set of rows, create DataTables for each section and pass to Panels.
        :param label_consolidation: pass 'object' to compare object equality for label consolidation, 'str'
                                    for converting all labels to strings then comparing equality. Use 'object'
                                    for more control over consolidation.
        :param enforce_label_order: pass False to allow consolidating lower labels even if upper labels do not match.
                                    e.g. if labels on one table are [['Top1'], ['Bot1', 'Bot2']], then labels on the other
                                    table are [['Top2'], ['Bot1', 'Bot2']], consolidated labels when passing False will be
                                    ['Bot1', 'Bot2'], while when passing True, no labels will be consolidated. Under True,
                                    will start from the top label, then stop consolidating once it has a mismatch.
        :param top_left_corner_labels: additional labels to place in the top left corner. pass a single string
                                       or a list of strings for convenience. a list of strings will be create labels
                                       which span the gap horizontally and go downwards, one label per row. pass
                                       LabelCollection or LabelTable for more control.
        :param pad_rows: horizontal spacing to put between panels
        :param pad_columns: vertical spacing to put between TableSections
        :param name: name that will be used to construct caption in output
        """
        self.name = name
        self.panels = panels
        self.label_consolidation = label_consolidation.lower().strip() \
            if isinstance(label_consolidation, str) else label_consolidation
        self.enforce_label_order = enforce_label_order
        self.top_left_corner_labels = _set_top_left_corner_labels(top_left_corner_labels)
        self.pad_rows = pad_rows
        self.pad_columns = pad_columns

        self.has_row_labels = False
        self.has_column_labels = False
        self.consolidate_labels()
        self.pad_grid()

    def iterpanels(self):
        """
        First panel is headers. Then each original panel

        self.grid includes all panels as well as labels. Need to separate back out to
        get each panel
        :return:
        :rtype:
        """

        if self.has_column_labels:
            orig_panel_index = -2
        else:
            orig_panel_index = -1
        name_used = False

        for row in self.rows:
            if not row.is_spacer: # don't increment original panel index when going through inserted spacing
                orig_panel_index += 1
                name_used = False # need to reset to use name again
            if orig_panel_index < 0 or name_used or self.panels[orig_panel_index].name is None:
                # column labels panel, no matching name. Or no name for user supplied panel
                yield Panel(PanelGrid([row]))
            else: # user passed panel, may have matching name
                full_name = panel_string(orig_panel_index) + self.panels[orig_panel_index].name
                yield Panel(PanelGrid([row]), name=full_name)
                name_used = True # only allow name to be used once

    @property
    def rows(self):
        try:
            return self._rows
        except AttributeError:
            self._rows = self._create_panel_rows()

        return self._rows

    @property
    def num_columns(self) -> int:
        try:
            return self._num_columns
        except AttributeError:
            self._num_columns = max([row.num_columns for row in self.rows])

        return self._num_columns

    def _create_panel_rows(self):
        rows: [TableSection] = []

        for panel_row in self.grid:
            new_row = None
            for i, section in enumerate(panel_row):
                if i == 0:
                    new_row = section
                else:
                    new_row = new_row + section
            if new_row:
                rows.append(new_row)

        try:
            num_columns = self._num_columns
        except AttributeError:
            num_columns = max([row.num_columns for row in rows])

        # Now pad rows
        for row in rows:
            row.pad(num_columns, direction='right')

        return rows


    @property
    def grid(self):
        try:
            return self._grid
        except AttributeError:
            self._grid = _concatenate_uneven_rows_filling_right(
                [panel.panel_grid for panel in self.panels],
                fill_value=TableSection([]),
                array_class=GridShape
            )

        return self._grid

    def consolidate_labels(self):

        if self.label_consolidation is None:
            return

        if self.label_consolidation == 'object':
            use_object_equality = True
        elif self.label_consolidation in ('string', 'str', True):
            use_object_equality = False
        else:
            raise ValueError(f'must pass object, string, or None to label consolidation. Got {self.label_consolidation}')

        column_labels: [LabelTable] = common_column_labels(
            self.grid,
            use_object_equality=use_object_equality,
            enforce_label_order=self.enforce_label_order
        )

        row_labels: [LabelTable] = common_row_labels(
            self.grid,
            use_object_equality=use_object_equality,
            enforce_label_order=self.enforce_label_order
        )

        if column_labels is not None:
            self._add_column_labels(column_labels)
            if row_labels is None:
                # If there are column labels but not row labels, still need to deal with top left label.
                # Adding the top left label is handled in the if row_labels is not None block, but it will
                # not be reached as row_labels is None. Therefore create a blank label table for each grid row
                # except for the column labels. The top left label to go with the column row will be added
                # in the following block.
                # Grid shape -1 to exclude just added column labels
                row_labels = [LabelTable([]) for row_index in range(self.grid.shape[0] - 1)]


        if row_labels is not None:
            # After adding column labels, there is an additional row at the top of the grid
            # Therefore we will need one additional LabelTable for the first row, which is the row of column labels
            # If top_left_corner_labels was passed on object creation, use that as LabelTable. Otherwise use a blank one
            if self.has_column_labels:
                all_row_labels = [self.top_left_corner_labels] + row_labels
            else:
                all_row_labels = row_labels
            self._add_row_labels(all_row_labels)

        # Remove from the original tables the labels that were just consolidated
        remove_label_collections_from_grid(
            self.grid,
            column_labels=column_labels,
            row_labels=row_labels,
            use_object_equality=use_object_equality
        )

    def pad_grid(self):
        row_pad = RowPadTable()

        grid_rows: [GridShape] = []
        for n_row, grid_row in enumerate(self.grid):
            # Add first elem
            new_row = np.array([grid_row[0]]).view(GridShape).reshape(1,1)
            # Add following elems
            for n_elem, elem in enumerate(grid_row[1:]):
                # Add pads between following elems
                if self.pad_columns and not (n_elem == 0 and self.has_row_labels): # only skip first if there are row labels
                    new_row = np.append(new_row, np.array([ColumnPadTable(self.pad_columns)])).view(GridShape)
                new_row = np.append(new_row, elem).view(GridShape)
                new_row = new_row.reshape((1, new_row.shape[0])) # reorganize into row
            grid_rows.append(new_row)
            # add row padding on every loop except last
            if n_row != self.grid.shape[0] - 1:
                # no need to add padding after column labels
                if n_row == 0 and self.has_column_labels:
                    continue
                for i in range(self.pad_rows):
                    grid_rows.append(
                        np.array([row_pad]).view(GridShape).reshape(1,1)
                    )

        # Before combining rows, must have same number of elements. Pad right with empty label tables
        num_grid_columns = max(row.shape[1] for row in grid_rows) # find max number of columns
        out_grid_rows = []
        for row in grid_rows:
            pad_number = num_grid_columns - row.shape[1]
            assert pad_number >= 0
            new_row = np.append(row, np.array([row_pad] * pad_number)).view(GridShape)
            new_row = new_row.reshape((1, new_row.shape[0]))  # reorganize into row
            out_grid_rows.append(new_row)

        new_grid = np.concatenate(out_grid_rows).view(GridShape)
        self._grid = new_grid
        self._rows = self._create_panel_rows() # need to recreate rows with new grid


    def _add_column_labels(self, column_labels: [LabelTable]):
        assert len(column_labels) == self.grid.shape[1]

        if all(table.is_empty for table in column_labels):
            # if no consolidated labels, no need to add
            self.has_column_labels = False
            return

        self.has_column_labels = True

        # Form PanelGrid from labels
        column_label_grid = PanelGrid(column_labels, shape=(1,len(column_labels)))

        # Combine label PanelGrid and existing PanelGrid
        self._grid = np.concatenate([column_label_grid, self._grid]).view(GridShape)

    def _add_row_labels(self, row_labels: [LabelTable]):
        assert len(row_labels) == self.grid.shape[0]

        if all(table.is_empty for table in row_labels):
            # if no consolidated labels, no need to add
            self.has_row_labels = False
            return

        self.has_row_labels = True

        # Form PanelGrid from labels
        row_label_grid = PanelGrid(row_labels, shape=(len(row_labels), 1))

        # Combine label PanelGrid and existing PanelGrid
        self._grid = np.concatenate([row_label_grid, self._grid], axis=1).view(GridShape)

    @classmethod
    def from_list_of_lists_of_dfs(cls, df_list_of_lists: [[pd.DataFrame]],
                                  panel_names: [str]=None, *args,
                                  panel_kwargs={},
                                  data_table_kwargs={}, **kwargs):
        """
        To create a single panel table, pass a single list within
        a list of DataFrames, e.g. [[df1, df2]] then shape will specify how the DataFrames will
        be organized in the Panel. If you pass two lists within the outer list, then shape will
        apply to each Panel. So [[df1, df2], [df3, df4]] with shape=(1,2) create a two Panel table
        with two tables placed within each panel going horizontally, so that the overall shape is (2,2).

        Note: convenience method for if not much control over table is needed.
        To apply different options to each panel, construct them individually using
        Panel.from_df_list

        :param df_list_of_lists:
        :param panel_names: list of panel names. Must be of same length as outer list in df_list_of_lists
        :param args: args to pass to PanelCollection constructor
        :param panel_kwargs: Panel.from_df_list kwargs. Same kwargs will be passed to all panels
        :param kwargs: kwargs to pass to PanelCollection constructor
        :param data_table_kwargs: kwargs to be passed to DataTable.from_df. Same kwargs will be passed to
                                  all data tables.

        :return: PanelCollection
        """
        _validate_panel_names(panel_names, df_list_of_lists)

        panels = []
        for i, df_list in enumerate(df_list_of_lists):
            panel_name = _panel_name_or_none(panel_names, i)
            panels.append(
                Panel.from_df_list(
                    df_list,
                    name=panel_name,
                    data_table_kwargs=data_table_kwargs,
                    **panel_kwargs)
            )

        label_consolidation = kwargs.pop('label_consolidation', 'string')

        return cls(
            panels,
            *args,
            label_consolidation=label_consolidation,
            **kwargs
        )

    def to_tex(self, mid_rule=True):
        from pyexlatex.table.logic.table.build import build_tabular_content_from_panel_collection
        return build_tabular_content_from_panel_collection(self, mid_rule=mid_rule)

def _panel_name_or_none(panel_names: [str], index: int):
    if panel_names is not None:
        return panel_names[index]
    else:
        return None

def _validate_panel_names(panel_names: [str], df_list_of_lists: [[pd.DataFrame]]):
    if panel_names is None:
        return

    num_panel_names = len(panel_names)
    num_panels = len(df_list_of_lists)
    if num_panel_names != num_panels:
        raise ValueError(f'must pass as many panel names as panels. Got {num_panel_names} names '
                         f'and {num_panels} panels.')


def _concatenate_uneven_rows_filling_right(rows, fill_value=np.nan, array_class=None):
    """
    Concatenates along vertical axis, filling right as needed.

    Examples:
        a = np.array(
            [[1, 2, 3]]
        )
        b = np.array(
            [[4, 5]]
        )
        _concatenate_uneven_rows_filling_right([a, b])

        array([[ 1.,  2.,  3.],
               [ 4.,  5., nan]])

    Args:
        rows:
        fill_value:
        array_class:

    Returns:

    """
    max_len = max([row.shape[1] for row in rows])

    concat_rows = []
    for row in rows:
        num_to_add = max_len - row.shape[1]
        if num_to_add > 0:
            add_array = np.array([[fill_value] * num_to_add])
            concat_array = np.concatenate([row, add_array], axis=1)
        else:
            concat_array = row
        concat_rows.append(concat_array)

    out_arr = np.concatenate(concat_rows)

    if array_class:
        out_arr = out_arr.view(array_class)

    return out_arr


from pyexlatex.models.item import MultiOptionSimpleItem
from pyexlatex.models.format.noindent import NoIndent


class Rule(MultiOptionSimpleItem):
    """
    Draw a horizontal line of given length and thickness
    """
    name = 'rule'

    def __init__(self, length: float = 2, thickness: float = 0.4):
        self.length = length
        self.thickness = thickness
        super().__init__(self.name, self.length_str, self.thickness_str)

    def __str__(self):
        contents = super().__str__()
        all_contents = NoIndent(contents)
        return str(all_contents)

    @property
    def length_str(self):
        if isinstance(self.length, float):
            return f'{self.length}cm'
        return self.length

    @property
    def thickness_str(self):
        if isinstance(self.thickness, float):
            return f'{self.thickness}pt'
        return self.thickness



from typing import Sequence, List, Union
from pyexlatex.presentation.beamer.templates.frames.grid import GridFrame, Frame
from pyexlatex.figure.models.graphic import Graphic
from pyexlatex.models.format.fills import VFill
from pyexlatex.presentation.beamer.templates.frames.dim_reveal import DimRevealMixin
from pyexlatex.models.sizes.textheight import TextHeight
from pyexlatex.models.format.centering import Centering
from pyexlatex.layouts.multicol import MultiCol


class TwoColumnFrame(GridFrame):
    """
    Creates a GridFrame, automatically setting widths and heights as an even split based on the shape
    of the content passed
    """

    def __init__(self, left_content, right_content, **frame_kwargs):
        super().__init__([[left_content, right_content]], **frame_kwargs)


class BasicTwoColumnFrame(Frame):
    """
    Similar to TwoColumnFrame but does not put any constraints on vertical sizing
    """

    def __init__(self, left_content, right_content, **frame_kwargs):
        self.content = [left_content, right_content]
        self.layout = MultiCol(self.content)
        super().__init__(self.layout.content, **frame_kwargs)


class TwoGraphicBase:

    def __init__(self, content, graphics: Sequence, graphics_on_right: bool = True, **frame_kwargs):
        self.content = content
        self.graphics = graphics
        self.graphics_on_right = graphics_on_right
        super().__init__(self.left_content, self.right_content, **frame_kwargs)

    @property
    def graphic_contents(self) -> List[Union[Graphic, VFill]]:
        graphic_contents = [Centering(), VFill()]
        for filepath_or_graphic in self.graphics:
            if isinstance(filepath_or_graphic, str):
                # Got filepath, need to make graphic
                graphic_contents.append(Graphic(
                    filepath_or_graphic,
                    width=0.9,
                    options=[
                        f'height={self.graphic_height}',
                        'keepaspectratio'
                    ]
                ))
            else:
                # Got a Graphic, TikZPicture, etc.
                graphic_contents.append(filepath_or_graphic)
            graphic_contents.append(VFill())
        graphic_contents.append(VFill())

        return graphic_contents

    @property
    def left_content(self):
        if self.graphics_on_right:
            return self.content
        else:
            return self.graphic_contents

    @property
    def right_content(self):
        if self.graphics_on_right:
            return self.graphic_contents
        else:
            return self.content

    @property
    def graphic_spacing(self) -> float:
        """
        Total amount of padding between images as a fraction of text height
        """
        return (len(self.graphics) - 1) * 0.20

    @property
    def graphic_height(self) -> str:
        total_area = 1 - self.graphic_spacing
        per_graphic = total_area / len(self.graphics)
        return f'{per_graphic}{TextHeight()}'


class TwoColumnGraphicFrame(TwoGraphicBase, TwoColumnFrame):
    """
    A GridFrame with graphics on one side and text on the other.
    """
    pass


class BasicTwoColumnGraphicFrame(TwoGraphicBase, BasicTwoColumnFrame):
    """
    Similar to TwoColumnGraphicFrame, but does not put any constraints on vertical sizing
    """
    pass


class TwoColumnGraphicDimRevealBase(DimRevealMixin):

    def __init__(self, content: Sequence[str], graphics: Sequence,
                 graphics_on_right: bool = True, ordered_list: bool = False,
                 **frame_kwargs):
        super().__init__(content, ordered_list=ordered_list,
                         graphics=graphics, graphics_on_right=graphics_on_right,
                         **frame_kwargs)


class TwoColumnGraphicDimRevealFrame(TwoColumnGraphicDimRevealBase, TwoColumnGraphicFrame):
    """
    A TwoColumnGraphicFrame where the non-graphic column is bulleted or numbered dim and reveal items
    """
    pass


class BasicTwoColumnGraphicDimRevealFrame(TwoColumnGraphicDimRevealBase, TwoColumnGraphicFrame):
    """
    A BasicTwoColumnGraphicFrame where the non-graphic column is bulleted or numbered dim and reveal items
    """
    pass


from typing import Optional, Sequence, List, Union
from pyexlatex.models.item import SimpleItem, ItemBase
from pyexlatex.models.credits.institution.inst import Inst
from pyexlatex.models.format.breaks import TableLineBreak
from pyexlatex.models.control.and_ import And


class Institutes(SimpleItem):
    """
    e.g. output
    \institute[VFU] % (optional)
    {
        \inst{1}%
        Faculty of Physics\\
        Very Famous University
        \and
        \inst{2}%
        Faculty of Chemistry\\
        Very Famous University
    }

    """
    name = 'institute'

    def __init__(self, institutions: Sequence[Sequence[str]], short_institution: Optional[str] = None):
        self.institutions = institutions
        self.short_institution = short_institution

        super().__init__(self.name, self.content_objs, pre_modifiers=self._wrap_with_bracket(short_institution))

    @property
    def content_objs(self) -> List[Union['Institution', And]]:
        # Eliminate duplicates in institutions, displaying unique ordered institutions
        count = 0
        counted_institutions = {}
        for inst_lines in self.institutions:
            inst_lines = tuple(inst_lines)
            if inst_lines not in counted_institutions:
                count += 1
                counted_institutions[inst_lines] = count

        content_objs = []
        for inst_lines, count in counted_institutions.items():
            content_objs.append(Institution(inst_lines, count))
            content_objs.append(And())
        content_objs = content_objs[:-1]  # strip last \and
        return content_objs


class Institution(ItemBase):
    """
    e.g.
        \inst{1}%
        Faculty of Physics\\
        Very Famous University
    """
    equal_attrs = [
        'institution_lines',
        'num'
    ]

    def __init__(self, institution_lines: Sequence[str], num: int):
        self.institution_lines = institution_lines
        self.num = num

    def __str__(self) -> str:
        from pyexlatex.logic.builder import _build
        inst = Inst(self.num)
        inst_output = TableLineBreak().join(self.institution_lines)
        return _build([
            inst,
            inst_output
        ])



from typing import Optional
from pyexlatex.typing import ListOrDictOrItem, StrListOrNone, AnyItem
from pyexlatex.logic.extract.docitems import is_latex_item


def get_attr_from_items_or_collection(content: ListOrDictOrItem, attr: str, unique: bool = False,
                                      collected_attr_values: Optional[list] = None) -> StrListOrNone:
    if unique:
        extend_func = _extend_with_items_not_in_list
    else:
        extend_func = _extend

    if collected_attr_values is None:
        collected_attr_values = []
        
    if isinstance(content, (list, tuple)):
        for item in content:
            get_attr_from_items_or_collection(item, attr, unique=unique, collected_attr_values=collected_attr_values)
    elif isinstance(content, dict):
        for name, item in content.items():
            get_attr_from_items_or_collection(item, attr, unique=unique, collected_attr_values=collected_attr_values)
    elif is_latex_item(content):
        if _is_item_and_has_attr(content, attr):
            extend_func(collected_attr_values, _get_from_item_or_item_data(content, attr))

    if not collected_attr_values:
        return None

    return collected_attr_values


def _is_item_and_has_attr(item: AnyItem, attr: str) -> bool:
    return is_latex_item(item) and _has_attr_or_data_attr_and_is_not_none(item, attr)


def _get_from_item_or_item_data(item: AnyItem, attr: str):
    if hasattr(item, attr):
        return getattr(item, attr)

    return getattr(item.data, attr)
    

def _extend_with_items_not_in_list(orig_list: list, extend_with: list) -> list:
    for item in extend_with:
        if item not in orig_list:
            orig_list.append(item)


def _extend(orig_list: list, extend_with: list) -> list:
    orig_list.extend(extend_with)


def _has_attr_and_is_not_none(item: AnyItem, attr: str) -> bool:
    return hasattr(item, attr) and getattr(item, attr) is not None


def _has_attr_or_data_attr_and_is_not_none(item: AnyItem, attr: str) -> bool:
    return _has_attr_and_is_not_none(item, attr) or (
        _has_attr_and_is_not_none(item, 'data') and _has_attr_and_is_not_none(item.data, attr)
    )


from pyexlatex.models.item import SimpleItem
from pyexlatex.models.package import Package


class Cite(SimpleItem):
    """
    Basic citation command, creates a reference to a document included in the bibliography.
    """
    name = 'cite'

    def __init__(self, item_accessor: str):
        super().__init__(self.name, item_accessor)


class NatBibCiteBase(Cite):

    def __init__(self, item_accessor):
        self.init_data()
        self.data.packages.append(Package('natbib'))
        return super().__init__(item_accessor)


class CiteT(NatBibCiteBase):
    """
    In text citation command, creates an in text (without parentheses) reference to a document
    included in the bibliography.
    """
    name = 'citet'


class CiteP(NatBibCiteBase):
    """
    Paragraph citation command, creates a paragraph (with parentheses) reference to a document
    included in the bibliography.
    """
    name = 'citep'



from typing import Sequence
from pyexlatex.layouts.base import LayoutBase
from pyexlatex.presentation.beamer.columns.column import Column
from pyexlatex.presentation.beamer.columns.columns import Columns

# TODO: move MultiCol to presentation as is only valid in beamer (verify this)

class MultiCol(LayoutBase):
    """
    Multi-column layout that does not make any constraints on vertical sizing.
    """

    def __init__(self, content: Sequence, **kwargs):
        self.orig_content = content
        super().__init__(content, **kwargs)
        self.content = self._get_column_content(self.contents)

    def _get_column_content(self, content: Sequence) -> Columns:
        num_cols = len(content)
        output = Columns(
            [Column(c, frac_of_page_height=None, frac_of_text_width=self.column_width) for c in content]
        )
        return output

    @property
    def column_width(self) -> float:
        num_cols = len(self.orig_content)
        spacing = (num_cols - 1) * 0.1
        split_width = 1 - spacing
        return split_width/num_cols




from typing import Sequence
from pyexlatex.models.item import ItemBase
from pyexlatex.models.section.base import TextAreaMixin


class Joined(TextAreaMixin, ItemBase):
    """
    Applies string.join on contents when output string is created, but retains content objects.
    """

    def __init__(self, contents, join_with=''):

        if not isinstance(contents, (list, tuple)):
            contents = [contents]

        self.orig_contents = contents
        self.join_with = join_with
        super().__init__(None, contents)

    def __str__(self):
        return self.join_with.join([str(content) for content in self.orig_contents])
